import asyncio
import logging

from nope.helpers import avgOfArray, ensureDottedAccess, maxOfArray, minOfArray
from nope.logger import getNopeLogger

logger = getNopeLogger('baseService', level=logging.INFO)


async def generatePingServices(dispatcher):
    async def ping():
        return ensureDottedAccess({
            'dispatcherId': dispatcher.id,
            'timestamp': dispatcher.connectivityManager.now
        })

    serviceName = 'nope/baseService/ping'

    await dispatcher.rpcManager.registerService(
        ping,
        ensureDottedAccess({
            'id': serviceName,
            'schema': {
                'inputs': [],
                'outputs': {
                    'type': 'object',
                    'properties': {
                        'dispatcherId': {
                            'type': 'string',
                            'description': 'Id of the responding Dispatcher'
                        },
                        'timestamp': {
                            'type': 'number',
                            'description': 'UTC-Timestamp of the system which is responding'
                        }
                    }
                }
            },
            'type': 'function',
            'description': 'Ping'
        })
    )

    logger.info("adding 'ping' service!")

    return await generatePingAccessors(dispatcher)


async def generatePingAccessors(dispatcher):
    serviceName = 'nope/baseService/ping'

    # Function to determine the ping in the services.
    async def determinePing(target: str):
        # Call the Pings
        start = dispatcher.connectivityManager.now
        result = await dispatcher.rpcManager.performCall(serviceName, [], {target})
        delay = dispatcher.connectivityManager.now

        ping = delay - start

        return ensureDottedAccess({
            "ping": ping,
            **result,
        })

    # Function to Ping all Services

    async def pingAll():
        dispatchers = list(
            dispatcher.rpcManager.services.keyMappingReverse.get(
                serviceName
            )
        )

        promises = []
        for target in dispatchers:
            promises.append(determinePing(target))

        if promises:
            pings = await asyncio.gather(*promises)

        _avg = avgOfArray(pings, "ping")
        _max = maxOfArray(pings, "ping")
        _min = minOfArray(pings, "ping")

        return ensureDottedAccess({
            "pings": pings,
            "avg": _avg,
            "max": _max,
            "min": _min,
        })

    return ensureDottedAccess({
        "determinePing": determinePing,
        "pingAll": pingAll
    })



from nope.dispatcher.core import NopeCore
from nope.helpers.pathMatchingMethods import comparePatternAndPath, ensureDottedAccess


class NopeDispatcher(NopeCore):
    """  A Dispatcher to perform a function on a Remote Dispatcher. Therefore a Task is created and forwarded to the remote.
    """

    @property
    def masterExists(self):
        return self.connectivityManager.master.isMasterForced

    def pushData(self, path: str, content, options=None) -> None:
        """ Pushs the desired data into the system.
        Args:
            path (str): The path to the Data.
            content (any): The Content to Push
            options (dict-like, optional): The Options during pushing. Defaults to None.
        """
        options = ensureDottedAccess(options)
        return self.dataDistributor.pushData(path, content, options)

    def pullData(self, path: str, _default=None):
        """ Helper to pull some data from the system.

        Args:
            path (str): The path to the Data.
            _default (any, optional): The value to use if no data has been found. If not provided an error is thrown. Defaults to None.

        Returns:
            any: The data or default type (if set)
        """
        return self.dataDistributor.pullData(path, _default)

    def subscribeToEvent(self, event: str, callback):
        """ Helper to subscribe to specific events.

        Args:
            event (str): Name of the event to listen to.
            callback (callable): the Callback to use. Must receive *args.

        Returns:
            Observer: Observer for the Subcription. contains the methods "pause", "unpause" and "unsubscribe"
        """
        return self.eventDistributor.registerSubscription(event, callback)

    def emitEvent(self, eventName, data, options=None):
        """ Emits an event with the given name. All event-subscriber, where the topic matches will receive this notification.

        Args:
            event (str): Name of the event to emii.
            data (any): The data to emit.
            options (dict-like): will be added to the event (for instance timestamp, sender, etc.)
        """
        options = ensureDottedAccess(options)
        self.eventDistributor.emit(eventName, data, options)

    def query(self, pattern: str, type: str) -> list:
        """ Receive the "instances" | "services" | "properties" | "events"
            which matches with the given pattern. Therefore the user provides
            the pattern and type.
            @return {string[]} List of the matching items.

        Args:
            pattern (str): pattern Pattern to query the provided type.
            type (str): type which should be querried. Allowed values are: "instances" | "services" | "properties" | "events"
        """

        items = []

        if type == 'instances':

            def callback_0(item):
                return item.identifier

            items = map(callback_0, self.instanceManager.
                        instances.data.getContent())
        elif type == 'services':
            items = list(self.rpcManager.services.simplified.keys())
        elif type == 'properties':
            items = self.dataDistributor.publishers.data.getContent()
        elif type == 'events':
            items = self.eventDistributor.publishers.data.getContent()
        else:
            raise Exception('Invalid Type-Parameter')

        return list(filter(lambda item: comparePatternAndPath(
            pattern, item).affected, items))

    def getAllHosts(self):
        hosts = set()
        for info in self.connectivityManager.dispatchers.originalData.values():
            hosts.add(info.host.name)
        return list(hosts)

    def toDescription(self):

        ret = self.connectivityManager.info.copy()
        ret.update(
            ensureDottedAccess({
                'isMaster': self.connectivityManager.isMaster,
                'instances': self.instanceManager.instances.data.getContent(),
                'services': self.rpcManager.services.data.getContent(),
                'events': self.eventDistributor.emitters,
                'properties': self.dataDistributor.emitters,
                'data': self.dataDistributor.pullData('', dict())
            })
        )

        return ret



from nope import PubSubSystem, NopeEventEmitter

called = 0

pub_sub = PubSubSystem()

publisher = NopeEventEmitter()
subscriber = NopeEventEmitter()

pub_sub.register(publisher, {
    "mode": "publish",
    "schema": {},
    "topic": "this/is/a/test",
})

pub_sub.register(subscriber, {
    "mode": "subscribe",
    "schema": {},
    "topic": "this/#",
})


def callback(data, rest):
    global called
    called += 1
    assert data == "Hello World"


def incremental(data, rest):
    print(data, rest)


pub_sub.onIncrementalDataChange.subscribe(incremental)

subscriber.subscribe(callback)
publisher.emit("Hello World")

assert called == 1



from nope.helpers import extractUniqueValues
from nope.merging import MergeData, DictBasedMergeData
from nope.observable import NopeObservable

called = 0


def callback(data, *args, **kwargs):
    global called
    called += 1


observable = NopeObservable()
observable.setContent(1)
sub = observable.subscribe(callback, {"skipCurrent": True})
observable.setContent(2)

d = {
    "a": {
        "a": [
            {
                "content": "a",
                "id": 1,
            },
            {
                "content": "b",
                "id": 2,
            },
        ], },
    "b": {
        "a": [
            {
                "content": "c",
                "id": 1,
            },
            {
                "content": "d",
                "id": 3,
            },
        ]
    }
}
res = extractUniqueValues(d, "a/+/content", "a/+/id")

d = dict()
merge = MergeData(d, lambda m: extractUniqueValues(m))

called = 0


def callback_01(data, *args, **kwargs):
    global called
    called += 1


merge.data.subscribe(callback_01)

d["a"] = "b"
d["b"] = "b"

merge.update()
merge.update()

r = merge.data.getContent()
assert called == 2, "Called the subscription twice"

d = {
    "a": {"key": "keyA", "data": "dataA"},
    "b": {"key": "keyB", "data": "dataB"}
}
merge = DictBasedMergeData(d, "data", "key")
merge.update()
print("H")



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de


from collections.abc import Iterable
from functools import reduce

from .dottedDict import DottedDict
from .objectMethods import rgetattr


def extractListElements(_list, _path: str):
    def extract(_prop):
        ret = rgetattr(_prop, _path)
        if ret:
            return ret

    return map(extract, _list)


def getElement(_list, operand, _path=''):
    for element in _list:
        if operand == rgetattr(element, _path):
            return element
    return None


def avgOfArray(_list, _path: str, defaultValue=0):
    if len(_list) == 0:
        return defaultValue

    values = map(lambda item: rgetattr(item, _path, defaultValue), _list)
    value = reduce(lambda prev, curr: prev + curr, values)
    return value / len(_list)


def minOfArray(_list, _path, defaultValue=float("inf")):
    if len(_list) == 0:
        return DottedDict({'min': defaultValue, 'index': None})

    values = list(map(lambda item: rgetattr(item, _path, defaultValue), _list))
    minValue = min(values)
    return DottedDict({'min': minValue, 'index': values.index(minValue)})


def maxOfArray(_list, _path, defaultValue=-float("inf")):
    if len(_list) == 0:
        return DottedDict({'max': defaultValue, 'index': None})

    values = list(map(lambda item: rgetattr(item, _path, defaultValue), _list))
    maxValue = max(values)
    return DottedDict({'max': maxValue, 'index': values.index(maxValue)})


def isIterable(obj):
    """ Helper to test if an objec is iterable.
    """

    if isinstance(obj, str):
        return False
    return isinstance(obj, Iterable)


def isList(obj) -> bool:
    """ Helper to check if the item is a list.

    Args:
        obj (any): The item to check

    Returns:
        bool: Result. If `True` => List
    """
    return type(obj) in (list,)


def flattenDeep(l: list) -> list:
    """ Helper to flatten a list of elements:

    Example:
        >>> l = [1, [2,3,[4,5]]]
        >>> flatten_deep(l)
        [1,2,3,4,5]

    Args:
        l (list): A nested list.

    Returns:
        list: A flattend list. (1 Dimension)
    """
    if len(l) == 0:
        return l
    if isinstance(l[0], list):
        return flattenDeep(l[0]) + flattenDeep(l[1:])
    return l[:1] + flattenDeep(l[1:])



from uuid import uuid4

from .stringMethods import replaceAll


def generateId(use_as_var: bool = False, pre_string: str = None) -> str:
    """ Helper to generate an id.
        if "use_as_var" is provided in the Options => the id is converted to an id
        if "pre_string" is provided, this string will be added in front of the id.


    Args:
        options (dotted_dict, optional):    The options. May cotains "use_as_var" and "pre_string".
                                            Defaults to dotted_dict({}).

    Returns:
        str: The id.
    """

    _id = str(uuid4())
    if use_as_var:
        _id = replaceAll(_id, '-', '')
        pre_string = pre_string if pre_string else '_'
    if pre_string:
        _id = pre_string + _id
    return _id



from .nopePubSubSystem import PubSubSystem
from ..helpers import copy, containsWildcards, ensureDottedAccess, MULTI_LEVEL_WILDCARD, flattenObject, \
    comparePatternAndPath


class DataPubSubSystem(PubSubSystem):

    def __init__(self, options={}):
        super().__init__(options)
        self._sendCurrentDataOnSubscription = True

    @property
    def data(self):
        # A Getter to return a COPY of the item. Outside of the system,
        # you'll never receive the original object.
        return copy(self._data)

    def pushData(self, path: str, content, options=None):
        """ Function, to push data. Every subscriber will be informed, if pushing the data on the
            given path will affect the subscribers.

        Args:
            path (str): The Path, on which the data should be changed
            content (any): The content to store in the given path.
            options (dict-like, optional): The Options, that will be forwarded to subscribers. Defaults to None.
        """
        options = ensureDottedAccess(options)
        return self._pushData(path, path, content, options)

    def pullData(self, path: str, default=None):
        """ Pull some Data of System. You will allways receive a just a copy. This method prevents you
            to use a pattern like path. If you want to use patterns please use the "patternbasedPullData"

        Args:
            path (str): The Path to pull the data.
            default (any, optional): If no data is found => return the default data. Defaults to None.

        Returns:
            any: The data.
        """

        return self._pullData(path, default)

    def patternbasedPullData(self, pattern: str, default=None):
        """ A Pattern based Pull. You can provide a mqtt based pattern and receive an array which contains
            all the data which matches the topic.

        Args:
            topic (str): The path to pull the data from
            default (any, optional): A Default value to use. If set to None we will skip this items. Defaults to None.

        Returns:
            list: the list containing an dict, with the keys: `path` = the matching path, `data` = the contained data.
        """
        return self._patternbasedPullData(pattern, default)

    def patternBasedPush(self, pattern: str, data, options=None, fast=False):
        """

        Args:
            pattern (str): The P
            data (any): _description_
            options (dict-like, optional): Additional Event-Date like `timestamp`, `sender`, ... . Defaults to None.
            fast (bool, optional):  Flag to emit single changes or all changes. Defaults to False.

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_
        """

        options = ensureDottedAccess(options)

        # To extract the data based on a Pattern,
        # we firstly, we check if we would affect the data.
        if not containsWildcards(pattern):
            return self.pushData(pattern, pattern, data, options)
        if MULTI_LEVEL_WILDCARD in pattern:
            raise Exception(
                'You can only use single-level wildcards in self action')
        flatten_data = flattenObject(self.data)
        options = self.updateOptions(options)
        for path in flatten_data.keys():
            if comparePatternAndPath(pattern, path).affectedOnSameLevel:
                self.pushData(path, pattern, data, options, fast)

        if fast:
            # Its better for us, to just store the incremental changes
            # with the pattern
            self.onIncrementalDataChange.emit(ensureDottedAccess(
                {'path': pattern, 'data': data, **options}))



from nope.helpers import generateId
from nope.logger import defineNopeLogger
from nope.observable import NopeObservable


class AbstractLayer:

    def __init__(self, uri: str, logger='info'):
        # Define the URI
        self.id = generateId()
        self.receivesOwnMessages = False

        self._logger = defineNopeLogger(logger, "core.mirror.io")

        self.connected = NopeObservable()
        self.connected.setContent(False)

    async def on(self, eventName: str, cb):
        raise Exception("Not implemented")

    async def off(self, eventName: str, cb):
        raise Exception("Not implemented")

    async def emit(self, eventName: str, data):
        raise Exception("Not implemented")

    async def dispose(self):
        raise Exception("Not implemented")

    def detailListeners(self, t, listeners):
        raise Exception("Not implemented")



# import nope
# from nope import getLayer, EXECUTOR, getDispatcher
# from nope.plugins import install, plugin

# """
# This is an Example how to extend the behavior of any class inside of Nope.

# Therefore we define a function (extend) which receives one Parameter, the
# module which is located under the given path. This module must contain
# the loaded class or function. We have to decorate the function with the
# ’plugin' decorator (which receives the path as parameter).

# Inside of the function we are able to access the original class using the
# module.*ACCESSOR*. In the example below we create an extra function
# "hello_dynamic" which will just give us a simple hello world response.

# Additional you are able to "store" functions into the module (see our
# extend function).

# If you want to manipulate specific vars --> please use a dict. for the
# key use the name of the variable you want to edit.
# """


# @plugin("nope.dispatcher.rpcManager")
# def extend_1(module):
#     "Extends `module` - rpcManager"

#     class NopeRpcManager(module.NopeRpcManager):
#         async def performCall(self, *args, **kwargs):
#             """ Extend the original behavior """
#             print("calling perform call - extend 1")
# return await module.NopeRpcManager.performCall(self, *args, **kwargs)

#         def hello_dynamic(self):
#             "Ask the plugin to say hello"
#             print("Hello from the Dynamic Plugin")

#     def hello_dynamic():
#         "Ask the plugin to say hello"
#         print("Hello from the Dynamic Plugin")

#     hello_dynamic_vars = {"hello": "Hello from the Dynamic Plugin"}
#     return NopeRpcManager, hello_dynamic, hello_dynamic_vars


# install(nope, extend_1, plugin_dest="nope.dispatcher.rpcManager")


# @plugin("nope.dispatcher.rpcManager")
# def extend_2(module):
#     "Extends `module` - rpcManager"

#     class NopeRpcManager(module.NopeRpcManager):
#         async def performCall(self, *args, **kwargs):
#             """ Extend the original behavior """
#             print("calling perform call - extend 2")
#             try:
#                 return await module.NopeRpcManager.performCall(self, *args, **kwargs)
#             except BaseException:
#                 print("Failed but we got the error")

#     return NopeRpcManager


# install(nope, [extend_1, extend_2], plugin_dest="nope.dispatcher.rpcManager")


# async def main():
#     dispatcher = getDispatcher({
#         "communicator": await getLayer("event"),
#         "logger": False,
#     })
#     manager = dispatcher.rpcManager
#     await dispatcher.ready.waitFor()

#     manager.hello_dynamic()
#     await manager.performCall("test", ["should be logged"])

#     # Now we are able to access our variable
#     print(nope.dispatcher.rpcManager.hello)

#     # and our new function, we injected into the module.
#     nope.dispatcher.rpcManager.hello_dynamic()


# EXECUTOR.loop.run_until_complete(main())



from .connectivy import generatePingServices
from ...helpers import ensureDottedAccess

SERVICES_NAME = {
    # "defineMaster": generateDefineMaster,
    "pingService": generatePingServices,
    # "timeSyncingService": enableTimeSyncing,
    # "syncingDataService": enablingSyncingData,
}


async def addAllBaseServices(dispatcher, opts=None):
    """ Adds Bases Services to the Dispatcher.

    Args:
        dispatcher (Nope): _description_
        opts (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """

    await dispatcher.ready.waitFor()
    services = ensureDottedAccess(None)

    if (opts and opts.services):
        for name in opts.services:
            services.update(await SERVICES_NAME[name](dispatcher))

    else:
        for name in SERVICES_NAME:
            services.update(await SERVICES_NAME[name](dispatcher))

    return services



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import sys
import traceback


def formatException(e):
    exception_list = traceback.format_stack()
    exception_list = exception_list[:-2]
    exception_list.extend(traceback.format_tb(sys.exc_info()[2]))
    exception_list.extend(traceback.format_exception_only(
        sys.exc_info()[0], sys.exc_info()[1]))

    exception_str = "Traceback (most recent call last):\n"
    exception_str += "".join(exception_list)
    # Removing the last \n
    exception_str = exception_str[:-1]

    return exception_str



from .dottedDict import ensureDottedAccess


class Emitter:
    """ A Event Emitter. This event emitter is used to listen to specific events and emit events.
        Therefore callbacks are registered with the method `on` and can be removed with `off`.
        An event is emitted using `emit`. To close the emitter -> no event will be forwarded

    """

    def __init__(self) -> None:
        """ Creates the Emitter.
        """
        self._subscribers = dict()
        self._paused = set()

    def on(self, event: str = None, callback=None):
        """ Adds a callback to the emitter. If fire event is emitted, the subscriber (callback) will be called.
        Args:
            event (str, optional): The event under which the callback is listening. Defaults to None.
            callback (callable, optional):  The callback to be called. Defaults to None.
                                            The Callback needs to receive `*args` and `**kwargs`

        Raises:
            TypeError: Raised, if the callback isnt callable.
        """
        if event not in self._subscribers:
            self._subscribers[event] = set()

        if not callable(callback):
            raise TypeError("The parameter 'callback', must be callable!")

        self._subscribers[event].add(callback)

        ret = ensureDottedAccess({
            'pause': lambda: self._pause(callback),
            'unpause': lambda: self._unpause(callback),
            'unsubscribe': lambda: self.off(event=event, callback=callback)
        })

        return ret

    def off(self, event: str = None, callback=None) -> bool:
        """ Removes the callback. Then it wont be called if an event has been emitted.

        Args:
            event (str, optional): The event under which the callback is subscribed. Defaults to None.
            callback (callable, optional): The Callback to remove. Defaults to None.
        Returns:
            bool:   True -> removed (has been present); False -> could not remove the callback (was not
                    subscribed for that event)
        """

        try:
            self._paused.remove(callback)
        except BaseException:
            pass

        if event in self._subscribers:
            try:
                self._subscribers[event].remove(callback)
                return True
            except BaseException:
                return False
        return False

    def _pause(self, callback):
        self._paused.add(callback)

    def _unpause(self, callback):
        try:
            self._paused.remove(callback)
        except BaseException:
            pass

    def emit(self, event: str = None, data=None, *args, **kwargs):
        """ Emits an event on with the given data.

        Args:
            event (str, optional): _description_. Defaults to None.
            data (optional): _description_. Defaults to None.
        """
        if event in self._subscribers:
            items = list(self._subscribers[event])
            for sub in items:
                if sub not in self._paused:
                    sub(data, *args, **kwargs)

    def close(self):
        """ Deletes all Subscribers.
        """
        self._subscribers = dict()
        self._paused = set()

    def amountOfSubscriptions(self, event=None) -> int:
        """ Returns, whether there are some subscribers listening.

        Returns:
            bool: _description_
        """

        if event is None:

            amount = 0

            for subs in self._subscribers.values():
                amount = amount + len(subs)

            return amount

        elif event in self._subscribers:
            return len(self._subscribers[event])
        return 0

    def hasSubscriptions(self, event=None) -> bool:
        return self.amountOfSubscriptions(event) > 0



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from ..eventEmitter import NopeEventEmitter
from ..helpers import DottedDict, ensureDottedAccess, generateId


class NopeObservable(NopeEventEmitter):
    """ An Observable storing the last value in `content`.
    """

    def __init__(self, options=None):

        if options is None:
            options = DottedDict()

        options.update({'showCurrent': True})
        super().__init__(options)
        self.id = generateId()
        self._value = None
        self._lastValue = None
        self._lastRest = None

    @property
    def emitter(self):
        return self._emitter

    def setContent(self, value, options=None):
        return self._emit(value, options)

    def _emit(self, value, options=None):
        options = ensureDottedAccess(options)
        if self.setter is not None and callable(self.setter):
            adapted = ensureDottedAccess(self.setter(value, options))
            if not adapted.valid:
                return False
            self._value = adapted.data
        else:
            self._value = value
        valueToPublish = self.getContent()
        if (not self.disablePublishing) and (
                options.forced or not (self._lastValue == valueToPublish)):
            return self._publish(valueToPublish, options)
        return False

    # def emit(self, *args, **kwargs):
    #     raise NotImplementedError("Not implemeneted on an Observable.")

    def _informSpecificObserver(self, observer):
        if self._lastRest is not None:
            # Call the last rest
            observer(ensureDottedAccess(
                {"value": self._lastValue, **self._lastRest}))

    def _publish(self, value, options=None):
        options = ensureDottedAccess(options)
        if options.forced or not self.disablePublishing:
            options = self._updateSenderAndTimestamp(options)
            self._lastRest = options
            self._lastValue = value
            self._emitter.emit(data=ensureDottedAccess(
                {'value': value, **options}))
            return self.hasSubscriptions
        return False

    def forcePublish(self, options=None):
        options = ensureDottedAccess(options)
        options.forced = True
        return self._publish(self.getContent(), options)

    def getContent(self):
        if self.getter is not None:
            return self.getter(self._value)
        return self._value

    @property
    def content(self):
        return self.getContent()

    def subscribe(self, callback, options=None):
        if options is None:
            options = {'mode': [
                'direct', 'sub', 'super'], 'skipCurrent': False}

        customCallback = self._adaptCallback(
            callback, ensureDottedAccess(options))

        res = self._subscribe(customCallback)

        # Because we are an observerable -> we will send the update:
        self._informSpecificObserver(customCallback)

        return res



import json
from logging import Logger
from socket import gethostname

import paho.mqtt.client as mqtt

from nope.helpers import replaceAll, generateId, formatException, SPLITCHAR
from nope.logger import getNopeLogger, INFO
from nope.observable import NopeObservable

HOSTNAME = gethostname()


def mqttMatch(subscription: str, offered: str):
    # Adapt the subscription and offered element to use the correct split char.
    subscription = replaceAll(subscription, SPLITCHAR, '/')
    offered = replaceAll(offered, SPLITCHAR, '/')

    # Perform the original method.
    res = mqtt.topic_matches_sub(subscription, offered)
    if res:
        return res

    if (len(offered.split('/')) > len(subscription.split('/'))
        ) and subscription.index('+') == 1:
        # Adapt the offered Topic, that is max as long as the subscription
        offered = ('/').join(offered.split('/')
                             [0:len(subscription.split('/'))])

        # Now we will use our default matcher again
        res = mqtt.topic_matches_sub(subscription, offered)

    elif (len(offered.split('/')) < len(subscription.split('/'))) and subscription.index('+') == 1:
        # Adapt the subscription Topic, that is max as long as the offered one
        subscription = ('/').join(subscription.split('/')
                                  [0:len(offered.split('/'))])

        # Now we will use our default matcher again
        res = mqtt.topic_matches_sub(subscription, offered)
    return res


class MQTTLayer:

    def __init__(self, uri, logger=INFO, preTopic=HOSTNAME,
                 qos=2, forwardToCustomTopics=True):

        self.uri = uri
        self.preTopic = preTopic
        self.qos = qos
        self.forwardToCustomTopics = forwardToCustomTopics
        self.uri = self.uri if self.uri.startswith(
            'mqtt://') else 'mqtt://' + self.uri
        self.connected = NopeObservable()
        self.connected.setContent(False)
        self._cbs = dict()
        self.logger: Logger = getNopeLogger(logger, 'core.layer.mqtt')
        self.considerConnection = True
        self.allowServiceRedundancy = False
        self._id = generateId()
        self.receivesOwnMessages = True

        # Create a mqtt client
        self._client = mqtt.Client(self.id, clean_session=False)

        # Create a handler for mqtt.
        def onConnect(*args, **kwargs):
            self.connected.setContent(True)

        def onDisconnect(*args, **kwargs):
            self.connected.setContent(False)

        def onMessage(client, user_data, msg):
            # call the messages
            self._onMessage(msg.topic, msg.payload.decode())

        self._client.onConnect = onConnect
        self._client.onDisconnect = onDisconnect
        self._client.onMessage = onMessage

        # Extract the Host and the Port based on the URI.
        _host = self.uri[len('mqtt://'):].split(":")[0]
        _port = self.uri[len('mqtt://'):].split(":")[1]

        self._client.connect(_host, int(_port))

        self._client.loop_start()

    @property
    def id(self):
        return self._id

    def _on(self, topic: str, callback):

        _topic = self._adaptTopic(topic)

        if _topic not in self._cbs:
            self._cbs[_topic] = set()
            self._logger.info("subscribing: " + _topic)
            self._client.subscribe(_topic, qos=self.qos)

        self._cbs[_topic].add(callback)

    def _off(self, topic: str, callback):
        """ Removes a callback from the item"""

        _topic = self._adaptTopic(topic)

        if _topic in self._cbs:
            self._cbs[_topic].remove(callback)

            if len(self._cbs[_topic]) == 0:
                self._logger.info("unsubscribing: " + _topic)
                self._client.unsubscribe(_topic)

    def _onMessage(self, topic: str, content):

        try:
            # the parsed data
            data = json.loads(content)

            for subscription in self._cbs:
                if mqttMatch(subscription, topic):
                    if self.logger:
                        self.logger.debug(
                            f'received message on "{topic}" with content={data}')

                    for cb in self._cbs[subscription]:
                        # perform the callback
                        cb(data)

        except Exception as E:
            self._logger.error(
                "Something went wrong during handling: '" + topic + "'. That shouldn't be the case")
            self._logger.error(formatException(E))

    def _emit(self, topic: str, data):

        _topic = self._adaptTopic()

        if not (_topic.startswith(HOSTNAME + "/nope/statusChanged")):
            self._logger.debug("emitting on " + _topic)
        self._client.publish(_topic, json.dumps(data), qos=self.qos)

    async def _on(self, eventName: str, cb):
        return self._on(f'+/nope/{eventName}', cb)

    async def emit(self, eventName: str, data):
        self._emit(f'{self.preTopic}/nope/{eventName}', data)

        if self.forwardToCustomTopics:
            if eventName == 'DataChanged':
                topic = data.path
                topic = self._adaptTopic(topic)
                await self.emit(topic, data.data)

            elif eventName == 'Event':
                topic = data.path
                topic = self._adaptTopic(topic)
                await self.emit(topic, data.data)

            elif eventName == 'RpcRequest':
                topic = data.functionId
                topic = self._adaptTopic(topic)
                await self.emit(topic, data.params)

    def _adaptTopic(self, topic: str):
        return replaceAll(topic, SPLITCHAR, '/')

    async def on(self, topic, callback):
        return self._on(topic, callback)

    async def off(self, topic, callback):
        return self._off(topic, callback)

    async def dispose(self):
        """ Kills the connection
        """
        self._client.disconnect()



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import json
import logging

from asyncio import sleep

from ..helpers import dynamicImport, formatException, ensureDottedAccess
from ..logger import getNopeLogger


def loadConfig(pathToFile: str):
    """ Load a config file.

        param: pathToFile: str File to load
    """
    config = {
        "functions": [],
        "packages": []
    }
    with open(pathToFile, 'r') as file:
        raw_data = file.read()
        config = json.loads(raw_data)

    return config


async def loadDesiredPackages(loader, packages_to_load, logger: logging.Logger = getNopeLogger('loadDesiredPackages'), delay=2):
    """ Load the Packages provided in the configuration
    """

    # Define the Return Array.
    pkgs = []

    # Scan for the Package-Files
    # And iterate over them.
    for item in packages_to_load:
        # Now Try to load a Package, to test whether is is an assembly.

        item = ensureDottedAccess(item)

        try:
            loadedPackage = dynamicImport(
                item["nameOfPackage"], item["path"]).DESCRIPTION

            loadedPackage = ensureDottedAccess(loadedPackage)

            loadedPackage["autostart"] = item["autostart"]
            loadedPackage["defaultInstances"] = item["defaultInstances"]

            # Add it to the list with elements.
            pkgs.append(loadedPackage)
        except Exception as e:
            if logger is not None:
                logger.error("Failed Loading the Package " +
                             item["nameOfPackage"])
                logger.error(formatException(e))
            else:
                print("Failed Loading the Package " +
                      item["nameOfPackage"])
                print(formatException(e))

    await loader.dispatcher.ready.waitFor()

    # Iterate over the Packages
    for pkgToLoad in pkgs:
        try:
            await loader.addPackage(pkgToLoad)
            if logger is not None:
                logger.info("Added Package " +
                            pkgToLoad["nameOfPackage"])

        except Exception as e:
            if logger is not None:
                logger.error("Failed add the Package " +
                             pkgToLoad["nameOfPackage"])
            else:
                print("Failed Loading the Package " +
                      item["nameOfPackage"])
                print(formatException(e))

    if (delay > 0):
        if logger:
            logger.info(f"Waiting ${delay} [s] before creating instances.")
        await sleep(delay)

    # Generate the instances.
    await loader.generateInstances()



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de


def generateAssignmentChecker(selector, core):
    # Generate a function which will return true.
    async def generate_true():
        return True

    if selector == 'first':
        return generate_true
    elif selector == 'cpu-usage':
        return generate_true
    elif selector == 'free-ram':
        return generate_true

    elif selector == 'dispatcher':
        async def dispatcher_matching(_module, used_dispatcher):
            return used_dispatcher.id == core.id

        return dispatcher_matching

    elif selector == 'host':
        host = core.connectivityManager.info.host.name

        async def host_matching(_module, used_dispatcher):
            return used_dispatcher.host.name == host

        return host_matching

    else:
        raise Exception('Please provide an valid selector')



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from nope.dispatcher.connectivityManager import NopeConnectivityManager
from nope.dispatcher.instanceManager import NopeInstanceManager
from nope.dispatcher.rpcManager import generateSelector, NopeRpcManager
from nope.helpers import ensureDottedAccess, generateId, EXECUTOR
from nope.logger import defineNopeLogger
from nope.observable import NopeObservable
from nope.pubSub import DataPubSubSystem, PubSubSystem


class NopeCore:

    def __init__(self, _options, id=None):

        _options = ensureDottedAccess(_options)

        def forwardEvent(item, rest):
            if item.sender != rcvExternally:
                EXECUTOR.callParallel(
                    self.communicator.emit,
                    'event',
                    ensureDottedAccess({
                        **item,
                        'sender': self._id
                    })
                )

        def forwardData(item, rest):
            if item.sender != rcvExternally:
                EXECUTOR.callParallel(
                    self.communicator.emit,
                    'dataChanged',
                    ensureDottedAccess({
                        **item,
                        'sender': self._id
                    })
                )

        def isReady(_):
            return self.connectivityManager.ready.getContent() \
                and self.rpcManager.ready.getContent() \
                and self.instanceManager.ready.getContent()

        def onEvent(msg):
            msg = ensureDottedAccess(msg)
            if msg.sender != self._id:
                data = msg.get("data")
                path = msg.get("path")
                msg.sender = rcvExternally
                self.eventDistributor.emit(path, data, msg)

        def onData(msg):
            msg = ensureDottedAccess(msg)
            if msg.sender != self._id:
                data = msg.get("data")
                name = msg.get("path")
                msg.sender = rcvExternally
                self.dataDistributor.pushData(name, data, msg)

        self._options = _options
        self._id = id

        self.communicator = _options.communicator

        if self._id is None:
            if _options.id:
                self._id = _options.id
            else:
                self._id = generateId()

        self.logger = defineNopeLogger(_options.logger, 'core')
        if self.logger:
            self.logger.info('setting up sub-systems.')

        self.eventDistributor = PubSubSystem()
        self.eventDistributor.id = self._id
        self.dataDistributor = DataPubSubSystem()
        self.dataDistributor.id = self._id

        defaultSelector = generateSelector(
            _options.get("defaultSelector", "first"), self)

        self.connectivityManager = NopeConnectivityManager(
            _options,
            self._id
        )
        self.rpcManager = NopeRpcManager(
            _options,
            defaultSelector,
            self._id,
            self.connectivityManager
        )
        self.instanceManager = NopeInstanceManager(
            _options,
            defaultSelector,
            self._id,
            self.connectivityManager,
            self.rpcManager,
            self
        )

        self.ready = NopeObservable()

        self.ready.getter = isReady
        rcvExternally = generateId()

        EXECUTOR.callParallel(
            self.communicator.on,
            'event',
            onEvent
        )

        self.eventDistributor.onIncrementalDataChange.subscribe(
            forwardEvent)

        EXECUTOR.callParallel(
            self.communicator.on,
            'dataChanged',
            onData
        )

        self.dataDistributor.onIncrementalDataChange.subscribe(
            forwardData)

        # Forward the Ready items.
        self.connectivityManager.ready.subscribe(
            lambda *args: self.ready.forcePublish())
        self.rpcManager.ready.subscribe(
            lambda *args: self.ready.forcePublish())
        self.instanceManager.ready.subscribe(
            lambda *args: self.ready.forcePublish())

        self.disposing = False

    @property
    def id(self):
        return self._id

    async def dispose(self):
        self.disposing = True
        self.ready.dispose()
        await self.eventDistributor.dispose()
        await self.dataDistributor.dispose()
        await self.connectivityManager.dispose()
        await self.rpcManager.dispose()
        await self.instanceManager.dispose()



import nope
from nope.plugins import install, plugin

"""
This is an Example how to extend the behavior of any class inside of Nope.

Therefore we define a function (extend) which receives one Parameter, the
module which is located under the given path. This module must contain
the loaded class or function. We have to decorate the function with the
’plugin' decorator (which receives the path as parameter).

Inside of the function we are able to access the original class using the
`module.*ACCESSOR*`. In the example below we create an extra function
"hello_dynamic" which will just give us a simple hello world response.

Additional you are able to "store" functions into the module (see our
extend function).

If you want to manipulate specific vars --> please use a dict. for the
key use the name of the variable you want to edit.
"""


@plugin("nope.helpers.timestamp")
def extend(module):
    "Extends `module` - rpcManager"

    def getTimestamp(options=None) -> str:
        return module.getTimestamp() + 500

    return getTimestamp


# Now install our plugin
nope, _, __ = install(nope, extend)

# The following main is just for clearification
print("ready")
print(nope.helpers.getTimestamp())
print("done")



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de
import asyncio
import itertools
import logging
import typing


from ..dispatcher import NopeDispatcher
from ..helpers import formatException, ensureDottedAccess
from ..logger import defineNopeLogger
from ..types import NopePackage


class NopePackageLoader():
    def __init__(self, dispatcher: NopeDispatcher, **options):
        _options = ensureDottedAccess(options)
        self._logger = defineNopeLogger(_options.log, 'core')
        self.packages: typing.Dict[str, NopePackage] = {}
        self.dispatcher = dispatcher
        self._instances = {}
        self._dispose_default_instance = []

    async def addPackage(self, package: NopePackage):
        """ Loader Function. self function will register all provided functions,
            create the desired instances. Additionally it will add all descriptors.

            param package: The Package to add.
        """
        if package.nameOfPackage in self.packages:
            raise Exception(
                "Already loaded a Package with the name \"" +
                package.nameOfPackage +
                "\" !"
            )

        if self._logger:
            self._logger.warn("loading package \"" +
                              package.nameOfPackage + "\"")

        # Store the Package:
        self.packages[package.nameOfPackage] = package

        # Based on the provided settings register a generator Function for the
        # Instances:
        for cl in package.providedClasses:
            # Get the Settings
            allowInstanceGeneration = cl.allowInstanceGeneration
            maxAmountOfInstance = cl.maxAmountOfInstance

            # Get the "create instance" function
            async def _not_provided(dispatcher, identifier):
                raise Exception(
                    "A function to create the function must be provided")

            createInstance = cl.createInstance
            selector = cl.selector

            if selector is not None and allowInstanceGeneration:
                # Create a Function that will generate functions
                async def _generate_instance(dispatcher, identifier):
                    currentAmount = self._instances.get(selector, 0)

                    if maxAmountOfInstance == -1 or currentAmount < maxAmountOfInstance:
                        # Define the Instance:
                        instance = await createInstance(dispatcher)
                        # Assign the Name
                        instance.identifier = identifier
                        # Update the Used Instance
                        self._instances[selector] = currentAmount + 1
                        # return the instance
                        return instance

                    raise Exception("Not allowed to create instances")

                await self.dispatcher.instanceManager.registerConstructor(
                    selector,
                    _generate_instance
                )

        for func in package.providedServices:
            if self._logger:
                self._logger.info("Package Loader generates the instances.")
            await self.dispatcher.rpcManager.registerService(func.function, func.options)

    async def generateInstances(self, test_requirements=True):
        """ Function to initialize all the instances.
        """
        if self._logger:
            self._logger.info("Package Loader generates the instances.")

        if test_requirements:
            # First extract all required Packages
            reuqiredPackages = set()

            for package in self.packages.values():
                reuqiredPackages |= set(package.requiredPackages)

            # Now check if the required packages are present or not.
            for req in reuqiredPackages:
                if req not in self.packages:
                    raise Exception("Packages are not known")

        # Now iterate over the Packages and define the defined instances.
        for package in self.packages.values():
            definitions = package.defaultInstances

            # Iterate over the Defined Instances.
            for definition in definitions:
                definition = ensureDottedAccess(definition)

                if self._logger:
                    self._logger.info("Requesting Generating Instance \"" +
                                      definition.identifier +
                                      "\" of type \"" +
                                      definition.type +
                                      "\"")

                instance = await self.dispatcher.instanceManager.createInstance(
                    definition
                )

                # Store the Function, that the instance will be disposed on
                # leaving.
                async def _dispose():
                    await instance.dispose()

                self._dispose_default_instance.append(_dispose)

                if self._logger:
                    self._logger.info("Generated Instance" +
                                      definition.identifier)

                # Perform the autostart:
                if definition.identifier in package.autostart:

                    if self._logger:
                        self._logger.info(
                            "Trying to perform autostart Instance" + definition.identifier)

                    try:
                        for task in package.autostart[definition.identifier]:
                            if task.delay > 0:
                                await asyncio.sleep(task.delay)

                            func = getattr(instance, task.name)
                            await func(*task.params)
                    except Exception as e:
                        if self._logger:
                            self._logger.error(
                                "Failed performing autostart for " + definition.identifier)
                            self._logger.error(formatException(e))
                        else:
                            print(
                                "Failed performing autostart for " +
                                definition.identifier)
                            print(formatException(e))
        if self._logger:
            self._logger.info("generated all defined Instances")



import asyncio
from asyncio import sleep

import pytest

from nope import getLayer, EXECUTOR, NopeConnectivityManager, DictBasedMergeData


@pytest.fixture
def event_loop():
    loop = asyncio.new_event_loop()
    EXECUTOR.assignLoop(loop)
    yield loop
    loop.close()


_mappingOfDispatchersAndServices = dict()
services = DictBasedMergeData(
    _mappingOfDispatchersAndServices,
    'services/+',
    'services/+/id')


async def get_manager(_communicator=None, _id=None):
    if _communicator is None:
        _communicator = await getLayer("event", "", False)

    manager = NopeConnectivityManager({
        'communicator': _communicator,
        "logger": False
    }, _id)

    return _communicator, manager


async def test_configuration():
    _, manager = await get_manager()

    await manager.ready.waitFor()

    await manager.setTimings({
        "checkInterval": 10,
        "dead": 25,
        "remove": 30,
        "sendAliveInterval": 5,
        "slow": 15,
        "warn": 20,
    })

    assert manager.ready.getContent(), "Manager not ready"

    assert manager.isMaster, "Didnt assing the communicator as master"
    manager.isMaster = True
    assert manager.isMaster, "Failed forcing the master"
    manager.isMaster = False
    assert not manager.isMaster, "Didnt removed the master flag"

    # Kill the manager.
    await manager.dispose()


async def main():
    _communicator, _first = await get_manager(None, "first")
    await _first.ready.waitFor()

    await sleep(0.1)

    _communicator, _second = await get_manager(_communicator, "second")
    await _second.ready.waitFor()

    # Wait for the first Handshake
    await sleep(1)

    assert _first.isMaster, "First should be master"
    _second.isMaster = True
    await sleep(0.01)
    assert not _first.isMaster, "Second should be master"

    _second.isMaster = None

    await sleep(0.01)
    assert _first.isMaster, "First should be master"

    print(_first.master)

    await _first.dispose()

    print("ready")


EXECUTOR.loop.run_until_complete(main())



from ..eventEmitter import NopeEventEmitter
from ..helpers import comparePatternAndPath, copy, generateId, DottedDict, ensureDottedAccess, rgetattr, \
    containsWildcards, getTimestamp, isIterable, rsetattr, flattenObject
from ..merging import DictBasedMergeData

DEFAULT_OBJ = object()


def _memoizedCompare(matchTopicsWithoutWildcards: bool):
    """ Helper to memoize the

    Args:
        options (bool): Flag to match topics with or without a wildcard

    Returns:
        callable: a Function that stores the results.
    """
    cache = {}

    def ret(pattern, path):
        nonlocal cache
        nonlocal matchTopicsWithoutWildcards

        key = f'{pattern}-{path}'

        if key not in cache:
            cache[key] = comparePatternAndPath(pattern, path, {
                'matchTopicsWithoutWildcards': matchTopicsWithoutWildcards
            })

        return cache[key]

    return ret


def _extractPubAndSubTopic(options: DottedDict):
    """ Helper to extract the pub and sub topic based on the provided options

    Args:
        options (DottedDict): the emitter options

    Returns:
        subTopic, pubTopic: the strings.
    """
    pubTopic = None
    if isinstance(options.topic, str):
        pubTopic = options.topic
    else:
        pubTopic = options.topic.publish

    subTopic = None
    if isinstance(options.topic, str):
        subTopic = options.topic
    else:
        subTopic = options.topic.subscribe

    return subTopic, pubTopic


class PubSubSystem:

    def __init__(self, mqttPatternBasedSubscriptions=True, forwardChildData=True,
                 forwardParentData=True, matchTopicsWithoutWildcards=True, **kwargs):

        # Adapt the Options
        self._options = ensureDottedAccess({
            'mqttPatternBasedSubscriptions': mqttPatternBasedSubscriptions,
            'forwardChildData': forwardChildData,
            'forwardParentData': forwardParentData,
            'matchTopicsWithoutWildcards': matchTopicsWithoutWildcards
        })

        self._sendCurrentDataOnSubscription = False
        self.id = generateId()

        self._emitters = dict()
        self._emittersToObservers = dict()

        self._matched = dict()
        self._disposing = False

        self._comparePatternAndPath = _memoizedCompare(self._options)
        self.subscriptions = DictBasedMergeData(self._emitters, 'subTopic')
        self.publishers = DictBasedMergeData(self._emitters, 'pubTopic')
        self.onIncrementalDataChange = NopeEventEmitter()

        self._data = ensureDottedAccess({}, useNoneAsDefaultValue=False)

    @property
    def options(self):
        return self._options.copy()

    def register(self, emitter, options):

        options = ensureDottedAccess(options)

        if options.topic is None:
            raise Exception("A Topic must be provided")

        if emitter not in self._emitters:
            # Now we will assign the pub and sub topic.
            # this is defined in the options.

            pubTopic = None
            if isinstance(options.topic, str):
                pubTopic = options.topic
            else:
                pubTopic = options.topic.publish

            subTopic = None
            if isinstance(options.topic, str):
                subTopic = options.topic
            else:
                subTopic = options.topic.subscribe

            if options.mode is None or not ("publish" in options.mode):
                pubTopic = False
            if options.mode is None or not ("subscribe" in options.mode):
                subTopic = False

            observer = None
            callback = None

            if pubTopic:

                def callbackToAssign(content, opts):
                    # Internal Data-Update of the pub-sub-system
                    # we wont push the data again. Otherwise, we
                    # risk an recursive endloop.
                    if opts.pubSubUpdate:
                        return

                    # We use this callback to forward the data into the system:
                    self._pushData(pubTopic, pubTopic,
                                   content, opts, False, emitter)

                callback = callbackToAssign

            # Register the emitter. This will be used during topic matching.
            self._emitters[emitter] = ensureDottedAccess(
                {'options': options, 'pubTopic': pubTopic, 'subTopic': subTopic, 'callback': callback,
                 'observer': observer})

            # Update the Matching Rules.
            self.updateMatching()

            if callback:
                # If necessary. Add the Callback.
                observer = emitter.subscribe(callback, ensureDottedAccess(
                    {'skipCurrent': not self._sendCurrentDataOnSubscription}))
                # Now lets store our binding.
                self._emittersToObservers[emitter] = observer

            if subTopic and self._sendCurrentDataOnSubscription:
                if containsWildcards(subTopic):

                    for item in self._patternbasedPullData(subTopic, None):
                        # We check if the content is available
                        if item.get("data", False):
                            emitter.emit(
                                item.data,
                                ensureDottedAccess({
                                    'sender': self.id,
                                    'topicOfContent': item["path"],
                                    'topicOfChange': item["path"],
                                    'topicOfSubscription': subTopic
                                })
                            )
                else:
                    currentContent = self._pullData(subTopic, None)
                    if currentContent is not None:
                        emitter.emit(
                            currentContent,
                            ensureDottedAccess({
                                'sender': self.id,
                                'topicOfContent': subTopic,
                                'topicOfChange': subTopic,
                                'topicOfSubscription': subTopic
                            })
                        )
        else:
            raise Exception('Already registered Emitter!')
        return emitter

    def updateOptions(self, emitter, options):
        # We will adapat the Options, for dotted access.
        options = ensureDottedAccess(options)
        if emitter in self._emitters:
            subTopic, pubTopic = _extractPubAndSubTopic(options)
            data = self._emitters.get(emitter)
            
            data.options = options
            data.subTopic = subTopic
            data.pubTopic = pubTopic

            self._emitters[emitter] = data
            
            self.updateMatching()
        else:
            raise Exception('Emitter is not registered')

    def unregister(self, emitter):
        if emitter in self._emitters:
            options = self._emitters.pop(emitter)
            subTopic, pubTopic = _extractPubAndSubTopic(options["options"])

            self._updatePartialMatching('remove', emitter, pubTopic, subTopic)
            return True
        return False

    def registerSubscription(self, topic, subscription):
        emitter = NopeEventEmitter()
        observer = emitter.subscribe(subscription)
        self.register(emitter, {'mode': 'subscribe',
                                'schema': {}, 'topic': topic})
        return observer

    @property
    def emitters(self):
        # TODO:
        pass

    def updateMatching(self):

        self._matched.clear()

        for item in self._emitters.values():
            # Extract the publisher topic
            pubTopic = item.get("pubTopic", False)

            if pubTopic is not False:
                self._updateMatchingForTopic(pubTopic)

        self.subscriptions.update()
        self.publishers.update()

    def _addMatchingEntryIfRequired(self, pubTopic, subTopic, emitter):
        result = self._comparePatternAndPath(subTopic, pubTopic)
        if result.affected:
            if not result.containsWildcards and ((result.affectedByChild and not self.options.forwardChildData) or
                                                 (
                result.affectedByParent and not self.options.forwardParentData)
            ):
                return

            # We now have match the topic as following described:
            # 1) subscription contains a pattern
            #    - dircet change (same-level) => content
            #    - parent based change => content
            #    - child based change => content
            # 2) subscription doesnt contains a pattern:
            #    We more or less want the data on the path.
            #    - direct change (topic = path) => content
            #    - parent based change => a super change
            if result.containsWildcards:
                if self.options.mqttPatternBasedSubscriptions:
                    if result.patternToExtractData is not False:
                        self._addToMatchingStructure(
                            'dataQuery', pubTopic, result.patternToExtractData, emitter)
                    elif result.pathToExtractData is not False:
                        self._addToMatchingStructure(
                            'dataPull', pubTopic, result.pathToExtractData, emitter)
                    else:
                        raise Exception(
                            'Implementation Error. Either the patternToExtractData or the pathToExtractData must be provided')
                else:
                    self._addToMatchingStructure(
                        'dataQuery', pubTopic, pubTopic, emitter)
            else:
                if result.affectedByChild and not self.options.forwardChildData or result.affectedByParent and not self.options.forwardParentData:
                    return
                if result.pathToExtractData is not False:
                    self._addToMatchingStructure(
                        'dataPull', pubTopic, result.pathToExtractData, emitter)
                else:
                    raise Exception(
                        "Implementation Error. The 'pathToExtractData' must be provided")

    def emit(self, eventName, data, options=None):
        return self._pushData(eventName, eventName, data,
                              ensureDottedAccess(options))

    async def dispose(self):
        self._disposing = True

        for emitter in self._emitters:
            self.unregister(emitter)

        self.onIncrementalDataChange.dispose()
        self.publishers.dispose()
        self.subscriptions.dispose()

    def _addToMatchingStructure(
            self, entry, topicOfChange, path_or_pattern, emitter):
        if topicOfChange not in self._matched:
            self._matched[topicOfChange] = ensureDottedAccess(
                {'dataPull': {}, 'dataQuery': {}})
        if path_or_pattern not in self._matched[topicOfChange][entry]:
            self._matched[topicOfChange][entry][path_or_pattern] = set()

        self._matched.get(topicOfChange)[entry].get(
            path_or_pattern).add(emitter)

    def _updateMatchingForTopic(self, topicOfChange):
        if topicOfChange not in self._matched:
            self._matched[topicOfChange] = ensureDottedAccess(
                {
                    'dataPull': {}, 
                    'dataQuery': {}
                }
            )

        # Find all matches
        for emitter, item in self._emitters.items():
            if item.subTopic is not False:
                self._addMatchingEntryIfRequired(
                    topicOfChange, item.subTopic, emitter)

    def _notify(self, topicOfContent: str, topicOfChange: str,
                options, emitterCausingUpdate=None):
        """ Internal Function to _notify all subscribers

        Args:
            topicOfContent (str): _description_
            topicOfChange (str): _description_
            options (dict-like): _description_
            _emitter (Emitter, optional): _description_. Defaults to None.
        """
        if self._disposing:
            return

        # Check whether a Matching exists for this
        # Topic, if not add it.
        if topicOfContent not in self._matched:
            self._updateMatchingForTopic(topicOfContent)

        referenceToMatch = self._matched[topicOfContent]

        for path_to_pull, emitters in referenceToMatch.dataPull.items():

            for emitter in emitters:
                # Only if we want to _notify an exclusive emitter we
                # have to continue, if our emitter isnt matched.
                if emitterCausingUpdate is not None and emitterCausingUpdate == emitter:
                    continue

                data = self._pullData(path_to_pull, None)

                emitter.emit(
                    data,
                    ensureDottedAccess({
                        **options,
                        'topicOfChange': topicOfChange,
                        'topicOfContent': topicOfContent,
                        'topicOfSubscription': self._emitters[emitter].subTopic
                    })
                )

        for pattern, emitters in referenceToMatch.dataQuery.items():
            
            # Get a copy of the filtered items.
            data = self._pullData(pattern, None)

            # Filter the items.
            data = filter(lambda item: self._comparePatternAndPath(topicOfChange, item.path).affected, data)

            if len(data) > 0:

                for emitter in emitters:
                    if emitter is not None and emitter != emitter:
                        continue
                    
                    emitter.emit(
                        data,
                        ensureDottedAccess({
                            **options,
                            'mode': 'direct',
                            'topicOfChange': topicOfChange,
                            'topicOfContent': topicOfContent,
                            'topicOfSubscription': self._emitters.get(emitter).subTopic
                        })
                    )

    def _updateOptions(self, options):
        if not options.timestamp:
            options.timestamp = getTimestamp()
        if not isinstance(options.forced, bool):
            options.forced = False
        if not isIterable(options.args):
            options.args = []
        if not options.sender:
            options.sender = self.id
        return options

    def _pushData(self, pathOfContent: str, pathOfChange: str,
                  data, options={}, quiet=False, emitter=None):
        """ Internal helper to push data to the data property. This results in informing the subscribers.

        Args:
            pathOfContent (str): Path, that is used for pushing the data.
            pathOfChange (str): The path, which caused the change.
            data (any): The data to push
            options (dict, optional): Options used during pushing. Defaults to {}.
            quiet (bool, optional): _description_. Defaults to False.
            emitter (_type_, optional): _description_. Defaults to None.
        """

        options = self._updateOptions(options)
        options.pubSubUpdate = True
        if containsWildcards(pathOfContent):
            raise 'The Path contains wildcards. Please use the method "patternbasedPullData" instead'
        elif pathOfContent == '':
            self._data = copy(data)
            self._notify(pathOfContent, pathOfChange, options, emitter)
        else:
            rsetattr(self._data, pathOfContent, copy(data))
            self._notify(pathOfContent, pathOfChange, options, emitter)
        if not quiet:
            self.onIncrementalDataChange.emit(ensureDottedAccess({
                'path': pathOfContent,
                'data': data,
                **options
            }))

    def _pullData(self, topic, default=None):
        if containsWildcards(topic):
            raise 'The Path contains wildcards. Please use the method "patternbasedPullData" instead'
        return copy(rgetattr(self._data, topic, default))

    def _patternbasedPullData(self, pattern: str, default=None):
        """ Helper, which enable to perform a pattern based pull.
            The code receives a pattern, and matches the existing content
            (by using there path attributes) and return the corresponding data.

        Args:
            pattern (str): the pattern
            default (any, optional): The default-value. Defaults to None.

        Returns:
            list:
        """

        # To extract the data based on a Pattern,
        # we firstly, we check if the given pattern
        # is a pattern.
        if not containsWildcards(pattern):
            # Its not a pattern so we will speed up
            # things.
            global DEFAULT_OBJ
            data = self._pullData(pattern, DEFAULT_OBJ)

            if data is not DEFAULT_OBJ:
                return [
                    ensureDottedAccess({
                        'path': pattern,
                        'data': data
                    })
                ]
            elif default is not None:
                return [
                    ensureDottedAccess({
                        'path': pattern,
                        'data': default
                    })
                ]

            return []

        # Now we know, we have to work with the query,
        # for that purpose, we will adapt the data object
        # to the following form:
        # {path: value}
        flatten = flattenObject(self._data)
        ret = []

        # We will use our alternative representation of the
        # object to compare the pattern with the path item.
        # only if there is a direct match => we will extract it.
        # That corresponds to a direct level extraction and
        # prevents to grap multiple items.
        for path, data in flatten.items():
            result = self._comparePatternAndPath(pattern, path)
            if result.affectedOnSameLevel or result.affectedByChild:
                ret.append(ensureDottedAccess({
                    'path': path,
                    'data': data
                }))

        # Now we just return our created element.
        return ret



import time

import socketio

from nope.helpers import generateId, EXECUTOR
from nope.logger import defineNopeLogger
from nope.observable import NopeObservable

PROFILE = False
OPEN_REQUESTS = {}


def profile_task(mode: str, data):
    try:
        if mode == "new":
            OPEN_REQUESTS[data["taskId"]] = time.process_time()
        elif mode == "done":
            if data["taskId"] in OPEN_REQUESTS:
                start = OPEN_REQUESTS.pop(data["taskId"])
                end = time.process_time()
                delta = end - start

                print(
                    "Bearbeitung hat",
                    round(
                        delta * 1000,
                        2),
                    "[ms] gedauert")
    except BaseException as e:
        print(f"Failed with mdoe= {mode}, data={data} cause: '{str(e)}'")


class IoSocketClientLayer:

    def __init__(self, uri: str, logger='info'):
        # Define the URI
        self.id = generateId()
        self.receivesOwnMessages = False
        self.uri = uri
        self.uri = self.uri if self.uri.startswith(
            'http://') else 'http://' + self.uri

        self._logger = defineNopeLogger(logger, "core.mirror.io")

        self._client = socketio.AsyncClient()
        EXECUTOR.callParallel(self._client.connect, self.uri)

        self.connected = NopeObservable()
        self.connected.setContent(False)

        def onConnect():
            self.connected.setContent(True)

        def onDisconnect():
            self.connected.setContent(False)

        self._client.on("connect", onConnect)
        self._client.on("disconnect", onDisconnect())

    async def on(self, eventName: str, cb):
        if PROFILE:
            if eventName == "rpcRequest":
                def req(data, *args):
                    profile_task("new", data)
                    cb(data)

                self._client.on(eventName, req)
            elif eventName == "rpcResponse":
                def res(data, *args):
                    profile_task("done", data)
                    cb(data)

                self._client.on(eventName, res)
            else:
                self._client.on(eventName, cb)
        else:
            self._client.on(eventName, cb)

    async def off(self, eventName: str, cb):
        self._client.handlers['/'].pop(eventName)

    async def emit(self, eventName: str, data):

        if PROFILE:
            if eventName == "rpcRequest":
                profile_task("new", data)
            if eventName == "rpcResponse":
                profile_task("done", data)

        await self._client.emit(eventName, data)

    async def dispose(self):
        await self._client.disconnect()

    def detailListeners(self, t, listeners):
        raise Exception('Method not implemented.')



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from copy import deepcopy

from .dottedDict import DottedDict, ensureDottedAccess
from .path import (MULTI_LEVEL_WILDCARD, SPLITCHAR, containsWildcards,
                   getLeastCommonPathSegment)
from .pathMatchingMethods import comparePatternAndPath

SENTINEL_1 = object()
SENTINEL_2 = object()


def rgetattr(data, path, default=SENTINEL_1, splitchar=SPLITCHAR):
    """ Helper to recursively get an value.

    Args:
        data (any): The data to extract the value from
        path (str): The path to extract the data from.
        default (any, optional): The Default value which will be return if no value has been found. If not provided None is returned
        splitchar (str, optional): _description_. Defaults to SPLITCHAR.

    Returns:
        any: The found data.
    """
    obj = data
    if len(path) > 0:
        if splitchar in path:
            for attr in path.split(splitchar):
                if type(obj) in (dict, DottedDict):
                    obj = obj.get(attr)
                else:
                    try:
                        obj = obj[attr]
                    except BaseException:
                        if default == SENTINEL_1:
                            return None
                        return default
        else:
            try:
                data = obj[path]
                return data
            except KeyError:
                if default == SENTINEL_1:
                    return None
                return default

    return obj


def rqueryAttr(data, query):
    if not containsWildcards(query):
        extractedData = rgetattr(data, query, SENTINEL_2)
        if extractedData == SENTINEL_2:
            return []

        return [
            DottedDict({
                'path': query,
                'data': extractedData
            })
        ]

    ret = []
    multi_level = MULTI_LEVEL_WILDCARD in query
    # Determine the max depth
    max_depth = None if multi_level else len(query.split(SPLITCHAR))

    # get the flatten object
    _dict = flattenObject(
        data,
        max_depth=max_depth,
        onlyPathToBaseValue=False
    )

    # Iterate over the items and use our path matcher to extract the matching
    # items.
    for iter_item in _dict.items():
        path = iter_item[0]
        value = iter_item[1]
        r = comparePatternAndPath(query, path)

        if r.affectedOnSameLevel or (multi_level and r.affectedByChild):
            ret.append(
                DottedDict({
                    'path': path,
                    'data': value
                })
            )

    return ret


def convertData(data, props):
    """ Helper to query data from an object.

        props is defined as followed:
        ```typescript
        props: {
            key: string;
            query: string;
        }[]
        ```

        Example 1:

        data = { "deep": { "nested": "test" } }
        result = convertData(data, [
            {
                "key": "result",
                "query": "deep/nested",
            },
        ])

        ==> result = [{"result": "test"}]

        Example 2:

        data = {
            "array": [
            {
                "data1": 0,
                "data2": "a",
            },
            {
                "data1": 1,
                "data2": "a",
            },
            ],
            "not": { "nested": "hello" }
        }

        result = convertData(data, [
            {
                "key": "a",
                "query": "array/+/data1",
            },
            {
                "key": "b",
                "query": "array/+/data2",
            },
        ])

        ==> result = [{"a": 0, "b": "a"}, {"a": 1, "b": "a"}]

    Args:
        data (any): The data
        props (dotted_dict[]): The query to use.
    """

    ret = DottedDict({})

    commonPattern = getLeastCommonPathSegment(
        list(
            map(
                lambda _item: _item.get("query"),
                props
            )
        )
    )

    for prop in props:
        ret[prop.get("key")] = rqueryAttr(data, prop.get("query"))

    helper = DottedDict({})

    for prop in props:
        items = ret[prop.get("key")]

        for idx, item in enumerate(items):

            if isinstance(commonPattern, str):
                result = comparePatternAndPath(commonPattern, item["path"])
                if result.pathToExtractData:
                    if not (result.pathToExtractData in helper):
                        helper[result.pathToExtractData] = DottedDict()
                    helper[result.pathToExtractData][prop.get(
                        "key")] = item.data
            else:
                if idx not in helper:
                    helper[idx] = DottedDict()

                helper[idx][prop.get("key")] = item.data

    return list(helper.values())


def _set(obj, accessor: str, value):
    try:
        obj[accessor] = value
    except BaseException as E:
        try:
            setattr(obj, accessor, value)
        except BaseException as E:
            raise E


def _get(obj, accessor, return_none=False):
    try:
        return obj[accessor]
    except BaseException as E:
        try:
            return getattr(obj, accessor)
        except BaseException as E:
            if return_none:
                return None
            raise E


def rsetattr(data, path: str, value, splitchar: str = SPLITCHAR):
    """ unction to Set recursely a Attribute of an Object

    Args:
        data (any):  The Object, where the data should be stored
        path (str): The Path of the Attribute. All are seprated by a the splitchar. (Defaults to'.' => For Instance 'a/b/0/a/c')
        value (any): The Value which should be Stored in the Attribute.
        splitchar (str, optional): The Splitchar to use. Defaults to "/". Defaults to SPLITCHAR.
    """

    obj = data

    ptrs = path.split(splitchar)

    for idx, attr in enumerate(ptrs[0:-1]):
        # Adapt the Object by going through a loop

        accessor = int(attr) if isInt(attr) else attr

        sub = _get(obj, accessor, return_none=True)

        if isinstance(obj, list):
            length = len(obj)
            while length <= accessor:
                obj.append(None)
                length = len(obj)

        if sub is None:
            # _obj is an Array and it doesnt contain the index
            # Extract the Next Element:
            nextAccessor = ptrs[idx + 1]

            nextIsInt = isInt(nextAccessor)

            if nextIsInt:
                _set(obj, accessor, [None] * (int(nextAccessor) + 1))
            else:
                _set(obj, accessor, DottedDict({}))

            sub = _get(obj, accessor)

        obj = sub

    try:
        obj[ptrs[len(ptrs) - 1]] = value
    except BaseException as E:
        try:
            sub = setattr(obj, ptrs[len(ptrs) - 1], value)
        except BaseException as E:
            raise E


def isInt(value) -> bool:
    """ Checks whether the Value is an Integer

    Args:
        value: Value to be checked
    """
    if isinstance(value, int):
        return True

    try:
        int(value)
        return True
    except BaseException:
        return False


def isFloat(value) -> bool:
    """ Checks whether the Value is a Float

    Args:
        value: Value to be checked
    """
    if isinstance(value, float):
        return True

    try:
        float(value)
        return True
    except BaseException:
        return False


def isNumber(value) -> bool:
    return isFloat(value) or isInt(value)


def objectToDict(obj):
    """ Function Converts a Object to a dict.

    Args:
        obj: The Object which should be converted.
    """

    ret = dict
    # Iterate through all properties of the Object
    for prop in obj:
        if not callable(obj):
            ret[prop] = obj[prop]

    return ret


def isDictLike(obj) -> bool:
    """ Check if the object is dict-like.

    Args:
        obj (any): The object to check

    Returns:
        bool:  Whether `obj` has dict-like properties.

    Examples:
        >>> isDictLike({1: 2})
        True
        >>> isDictLike([1, 2, 3])
        False
        >>> isDictLike(dict)
        False
        >>> isDictLike(dict())
        True
    """
    attributes_to_check = ("__getitem__")

    for attr in attributes_to_check:
        if not hasattr(obj, attr):
            return False

    # Prevent returning Classes:
    return not isinstance(obj, type)


def allowsSubscripton(obj):
    return hasattr(obj, "__getitem__")


def flattenObject(data, prefix="", splitchar=SPLITCHAR,
                  onlyPathToBaseValue=False, max_depth=None):
    ret = DottedDict()

    # Check if we are ae to access the items using an subscription.
    if allowsSubscripton(data):
        def callback(path, data, *args):
            ret[path] = data

        recursiveForEach(
            data,
            callback,
            prefix,
            splitchar,
            onlyPathToBaseValue,
            max_depth
        )

    return ret


def deflattenObject(flattenObject, options=None):
    _options = ensureDottedAccess({'prefix': '', 'splitchar': SPLITCHAR})
    _options.update(ensureDottedAccess(options))

    ret = ensureDottedAccess({})

    for key, val in flattenObject.items():
        if _options.prefix != '':
            key = key[len(_options.prefix):]
        rsetattr(ret, key, val, _options.splitchar)

    return ret


def getKeys(obj):
    """ Helper, to get keys of an array, dict etc.

    Args:
        obj: The considered Object.

    Returns:
        list: List of Keys
    """
    keys = list()
    t_obj = type(obj)
    if not (t_obj is str) and not (callable(obj)):
        if t_obj in (list, set):
            # For list, we will use the index
            keys = range(0, len(obj))
        elif hasattr(obj, "keys"):
            # We are working with a dict.
            keys = list(obj.keys())
        else:
            try:
                # For all other items, we will use the vars as key.
                keys = list(vars(obj))
            except BaseException:
                pass

    return keys


def deepAssign(target, source):
    """ Deeply assigns the items given in the dict, whereas the
        keys of the source will be used as path, its value as value
        to assign.

    Args:
        target (any): original object.
        source (dict | dict-like): Todo

    Returns:
        any: the manipulated target
    """
    flattend = flattenObject(source)
    for iter_item in flattend.entries():
        path = iter_item[0]
        value = iter_item[1]
        rsetattr(target, path, value)
    return target


def recursiveForEach(obj, callback=None, prefix="", splitchar=SPLITCHAR, callOnlyOnBaseValues=True,
                     max_depth=None, parent='', level=0):
    """ Function, that will iterate over an object.

    Args:
        obj: The Object to iterate
        prefix (_type_): A prefix for the Path.
        callback (callable, optional): Callback, that will be called.. Defaults to None.
        splitchar (_type_, optional): The Splitchar to use, to generate the path. Defaults to SPLITCHAR.
        callOnlyOnBaseValues (bool, optional): A Flag, to call the Function only on Values. Defaults to True.
        max_depth (_type_, optional): Determine the max Depth, after which the Iteration will be stopped.. Defaults to None.
        parent (str, optional): For Recursive call only. Defaults to ''.
        level (int, optional): For Recursive call only. Defaults to 0.
    """
    if isInt(max_depth) and level > max_depth:
        # Break the Loop
        return

    # No try to extrac the Keys.
    keys = getKeys(obj)

    called = False
    if not callOnlyOnBaseValues and callable(callback):
        callback(prefix, obj, parent, level)
        called = True

    # If there exits some keys:
    if len(keys) > 0:
        # We will iterate over the and
        for key in keys:
            # Define the variable, containing the path
            path = str(key) if '' == prefix else prefix + splitchar + str(key)

            if obj[key] is not None:
                if hasattr(obj[key], "to_json") and callable(obj[key].to_json):
                    data = obj[key].to_json()
                    recursiveForEach(
                        data,
                        callback,
                        path,
                        splitchar,
                        callOnlyOnBaseValues,
                        max_depth,
                        prefix,
                        level + 1
                    )

                else:

                    recursiveForEach(
                        obj[key],
                        callback,
                        path,
                        splitchar,
                        callOnlyOnBaseValues,
                        max_depth,
                        prefix,
                        level + 1
                    )
    elif not called:
        callback(prefix, obj, prefix, level)


def keepProperties(obj, properties):
    if allowsSubscripton(obj):
        ret = DottedDict({})
        default_obj = DottedDict({'error': True})

        for key in getKeys(obj):
            value = rgetattr(obj, key, default_obj)
            valueToAssign = value
            if value != default_obj:
                valueToAssign = copy(value)
            else:
                valueToAssign = properties[key]()
            rsetattr(ret, key, valueToAssign)

        return ret
    raise Exception('Function can only create Objects')


WARNED = False


def copy(obj):
    """ A Helper, which can be used to receive a copy.

    Args:
        obj (any): The Object to copy

    Returns:
        bool, any: A flag, to show whether it succeeded or failed; the copy (or if failed the object) itself.
    """
    try:
        return deepcopy(obj)
    except BaseException:
        try:
            return obj.copy()
        except BaseException:
            global WARNED
            if not WARNED:
                WARNED = True
                print("Failed to create a copy using orignal value.")
            return obj



from dataclasses import dataclass, field
from typing import Any, Callable, List, Dict, Union


@dataclass
class ServiceUi:
    """ Definition of an ui.

        parameters:
            file =                          An additional file (javascript) -> defining the ui. Defaults to False -> No extra File is required.
            autoGenBySchema =               Autogenerates the ui to configure the service based on the given schema
            requiredProvidersForRendering = Defined items (service), that must be present in the network, to define the ui-file.
    """

    file: str | bool = False
    """ An additional file (javascript) -> defining the ui. Defaults to False -> No extra File is required.
    """

    autoGenBySchema: bool = False
    """ Autogenerates the ui to configure the service based on the given schema
    """

    requiredProvidersForRendering: List[str] = field(default_factory=list)
    """ Defined items, that must be present in the network,
        to define the ui-file.
    """


@dataclass
class CallOptions:
    """ Options used in an RPC-Call

        parameters:
            resultSink  = Desired result sink. If implemented, the result will be published on this topic as well.
            timeout     = A User Provided Timeout of the call. After the timeout has been ellapsed, the task is
                          cancelled with a timeout error. The Time is given in **ms**
    """
    resultSink: str = None
    """ Desired result sink. If implemented, the result will be published on this topic as well. """

    timeout: int = -1
    """ A User Provided Timeout of the call. After the timeout has been ellapsed, the task is
        cancelled with a timeout error. The Time is given in **ms**
    """


@dataclass
class ServiceOptions:

    schema: Any
    """ Schmea describing the Function.
    """

    resultSink: str = None
    """ Desired result sink. If implemented, the result will be published on this topic as well. """

    timeout: int = -1
    """ A User Provided Timeout of the call. After the timeout has been ellapsed, the task is
        cancelled with a timeout error. The Time is given in **ms**
    """

    id: str | bool = False
    """ Instead of generating a uuid an id could be provided
    """

    ui: ServiceUi = field(default_factory=ServiceUi)
    """ Defintion of the UI.
    """

    isDynamic: bool = False
    """ Flag to mark the Function as Dynamically created
    """


@dataclass
class ProvidedClass:
    selector: str
    createInstance: Callable[Any, Any]
    allowInstanceGeneration: bool = True
    maxAmountOfInstance: int = -1


@dataclass
class DefaultInstance:
    identifier: str
    type: str
    params: List[Any] = field(default_factory=list)


@dataclass
class ProvidedService:
    function: Callable[Any, Any]
    options: ServiceOptions


@dataclass
class Autostart:
    name: str
    params: List[Any] = field(default_factory=list)
    delay: int = -1


@dataclass
class NopePackage:
    nameOfPackage: str
    autostart: Dict[str, List[Autostart]] = field(default_factory=dict)
    defaultInstances: List[DefaultInstance] = field(default_factory=list)
    providedServices: List[ProvidedService] = field(default_factory=list)
    providedClasses: List[ProvidedClass] = field(default_factory=list)
    requiredPackages: List[str] = field(default_factory=list)



import os
import os
import platform
import re
import subprocess
from enum import IntEnum
from socket import gethostname

import psutil

from nope.helpers import getTimestamp, ensureDottedAccess, generateId, DottedDict, minOfArray, formatException, \
    EXECUTOR
from nope.logger import defineNopeLogger
from nope.merging import DictBasedMergeData
from nope.observable import NopeObservable


class ENopeDispatcherStatus(IntEnum):
    HEALTHY = 0
    SLOW = 1
    WARNING = 2
    DEAD = 3


def _get_processor_name():
    if platform.system() == "Windows":
        return platform.processor()
    elif platform.system() == "Darwin":
        os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
        command = "sysctl -n machdep.cpu.brand_string"
        return subprocess.check_output(command).strip()
    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = subprocess.check_output(
            command, shell=True).strip().decode()
        for line in all_info.split("\n"):
            if "model name" in line:
                return re.sub(".*model name.*:", "", line, 1)
    return ""


_PROCESSOR_NAME = _get_processor_name()

# We want to load the file containing the current version.
_VERSION = "1.4.1"
try:
    _PATH = os.path.join(os.path.dirname(__file__), "..", "..", "VERSION")
    _VERSION = open(_PATH).read().strip()
except BaseException:
    pass


class NopeConnectivityManager:

    def __init__(self, options, id=None):

        options = ensureDottedAccess(options, False)

        self.options = options
        self._id = id if id is not None else generateId()
        self._deltaTime = 0
        self._checkStatusTask = None
        self._sendStatusTask = None

        self._timeouts = ensureDottedAccess({})

        self._communicator = options.communicator
        self._connectedSince = getTimestamp()
        self._isMaster = options.isMaster if isinstance(
            options.isMaster, int) else None

        self._logger = defineNopeLogger(
            options.logger, 'core.connectivity-manager')

        self.ready = NopeObservable()
        self.ready.setContent(False)
        self._externalDispatchers = dict()
        self.dispatchers = DictBasedMergeData(self._externalDispatchers, 'id')

        if self._logger:
            self._logger.info(f'core.connectivity-manager {self.id} is ready')

        self.reset()
        EXECUTOR.callParallel(self.init)

    @property
    def id(self):
        return self._id

    @property
    def info(self):
        return self._info()

    def _info(self):

        # Helper for the Memory.
        _virtual_memory = psutil.virtual_memory()

        return ensureDottedAccess({
            'id': self.id,
            'env': 'python',
            'version': _VERSION,
            'isMaster': self.isMaster,
            'isMasterForced': isinstance(self._isMaster, bool),
            'host': {
                'cores': os.cpu_count(),
                'cpu': {
                    'model': _PROCESSOR_NAME,
                    'speed': psutil.cpu_freq().max,
                    'usage': psutil.cpu_percent(0)
                },
                'os': str(platform.system()) + " " + str(platform.release()),
                'ram': {
                    'usedPerc': _virtual_memory.percent,
                    'free': round(_virtual_memory.free / 1048576),
                    'total': round(_virtual_memory.total / 1048576)
                },
                'name': gethostname()
            },
            'pid': os.getpid(),
            'timestamp': self.now,
            'connectedSince': self.connectedSince,
            'status': ENopeDispatcherStatus.HEALTHY.value,
            'plugins': []
        })

    @property
    def upTime(self):
        return getTimestamp() - self.connectedSince

    @property
    def connectedSince(self):
        return self._connectedSince + self._deltaTime

    @property
    def isMaster(self):
        if self._isMaster is None:
            try:
                return self.id == self.master.id
            except Exception as e:
                return False
        return self._isMaster

    @isMaster.setter
    def isMaster(self, value):
        self._isMaster = value
        EXECUTOR.callParallel(self._asyncSendStatus)

    def _getPossibleMasterCandidates(self):
        possibleMasters = []
        for info in self.dispatchers.originalData.values():
            # In here we have to use Camel-Case because we are considering
            # Camel-Case in the entire system.
            if info.isMasterForced and not info.isMaster:
                continue
            possibleMasters.append(info)
        return possibleMasters

    @property
    def master(self):
        def checkIfIsMaster(item):
            return item.isMaster and item.isMasterForced

        candidates = self._getPossibleMasterCandidates()
        masters = list(filter(checkIfIsMaster, candidates))
        if len(masters) != 1:
            if len(masters) > 1 and self._logger:
                self._logger.warn(
                    f"Found {len(masters)}. We now will select the one which has been online for the longest time.")
            idx = minOfArray(candidates, 'connectedSince').index
            if idx is not None:
                return candidates[idx]
            raise Exception('No Master has been found !')
        return masters[0]

    @property
    def now(self):
        return getTimestamp() + self._deltaTime

    async def init(self):

        self.ready.setContent(False)

        await self.setTimings(self.options.timeouts)

        def onConnect(connected, _rest):
            if connected:
                self._connectedSince = getTimestamp()
                EXECUTOR.callParallel(self.emitBonjour)
                EXECUTOR.callParallel(self._asyncSendStatus, forced=True)

        self._communicator.connected.subscribe(onConnect)
        await self._communicator.connected.waitFor()

        def onStatusChanged(info):
            self._externalDispatchers[info.id] = ensureDottedAccess(info)
            if info.id != self.id:
                self._externalDispatchers[self.id] = self.info
                self.dispatchers.update()

        await self._communicator.on('statusChanged', onStatusChanged)

        def onBonjour(opts):
            if self.id != opts.dispatcherId:
                if self._logger:
                    self._logger.debug(
                        'Remote Dispatcher "' + opts.dispatcherId + '" went online')
                # Ensure Sending our Message
                EXECUTOR.callParallel(self._asyncSendStatus, forced=True)

        await self._communicator.on('bonjour', onBonjour)

        def onAurevoir(msg):
            # We try to pop the item. If it fails we wont update the elements
            item = self._externalDispatchers.pop(msg.dispatcherId, False)
            if item:
                self.dispatchers.update()

        await self._communicator.on('aurevoir', onAurevoir)

        await self.emitBonjour()
        await self._asyncSendStatus(forced=True)

        if self._logger:
            self._logger.info(
                f'core.connectivity-manager {self.id} is initialized')

        self.ready.setContent(True)

    def _checkDispachterHealth(self):
        """ Checks the health of dispatcher. If some changes like their connection
            are determined, an update is transmitted via the attribute `dispatchers`

        """

        # If no other Dispatcher is present,
        # This test is Obsolete
        if len(self._externalDispatchers) <= 1:
            return

        currentTime = self.now

        changes = False

        for status in list(self._externalDispatchers.values()):
            # determine the Difference
            diff = currentTime - status['timestamp']

            # Based on the Difference Determine the Status
            if diff > self._timeouts['remove']:
                # remove the Dispatcher. But be quiet.
                # Perhaps more dispatchers will be removed
                self._removeDispatcher(status['id'], True)
                changes = True
            elif diff > self._timeouts['dead'] and status['status'] != ENopeDispatcherStatus.DEAD:
                status['status'] = ENopeDispatcherStatus.DEAD
                changes = True
            elif self._timeouts['warn'] < diff <= self._timeouts['dead'] and status[
                    'status'] != ENopeDispatcherStatus.WARNING:
                status['status'] = ENopeDispatcherStatus.WARNING
                changes = True
            elif self._timeouts['slow'] < diff <= self._timeouts['warn'] and status[
                    'status'] != ENopeDispatcherStatus.SLOW:
                status['status'] = ENopeDispatcherStatus.SLOW
                changes = True
            elif diff <= self._timeouts['slow'] and status['status'] != ENopeDispatcherStatus.HEALTHY:
                status['status'] = ENopeDispatcherStatus.HEALTHY
                changes = True

        if changes:
            # Execute the following function parallel.
            EXECUTOR.callParallel(lambda: self.dispatchers.update)

    def _removeDispatcher(self, dispatcher: str, quiet=False):
        """ Removes a dispatcher.
        """
        dispatcherInfo = self._externalDispatchers.pop(dispatcher, None)
        if not quiet:
            self.dispatchers.update()
        if self._logger and dispatcherInfo:
            self._logger.warn(
                f'a dispatcher on {dispatcherInfo.host.name} went offline. ID of the Dispatcher: "{dispatcher}"')

    async def _asyncSendStatus(self, forced=False):
        if self._communicator.connected.getContent():
            try:
                info = self.info
                self._externalDispatchers[self.id] = info

                if forced or len(self._externalDispatchers) > 1:
                    await self._communicator.emit('statusChanged', info)
            except Exception as e:
                if self._logger:
                    self._logger.error('Failled to send the status')
                    self._logger.error(formatException(e))

                else:
                    print(formatException(e))

    def syncTime(self, timestamp, delay=0):
        internalTimestamp = getTimestamp()
        self._deltaTime = internalTimestamp - (timestamp - delay)

    def getStatus(self, _id: str):
        return self._externalDispatchers[_id]

    async def emitBonjour(self):
        await self._communicator.emit('bonjour',
                                      ensureDottedAccess({'dispatcherId': self.id}))

    def reset(self):
        self._externalDispatchers.clear()
        self.dispatchers.update(self._externalDispatchers)

    async def setTimings(self, options):

        options = ensureDottedAccess(options)

        await self.dispose(True)

        self._timeouts = ensureDottedAccess({
            'sendAliveInterval': 500,
            'checkInterval': 250,
            'slow': 1000,
            'warn': 2000,
            'dead': 5000,
            'remove': 10000
        })

        if options:
            self._timeouts.update(options)

        # Setup Test Intervals:
        if self._timeouts.checkInterval > 0:
            # Define a Checker, which will test the status
            # of the external Dispatchers.
            self._checkStatusTask = EXECUTOR.setInterval(
                self._checkDispachterHealth,
                self._timeouts["checkInterval"]
            )

        if self._timeouts.sendAliveInterval > 0:
            # Define a Timer, which will emit Status updates with
            # the desired delay.
            self._sendStatusTask = EXECUTOR.setInterval(
                self._asyncSendStatus,
                self._timeouts["sendAliveInterval"]
            )

    @property
    def allHosts(self):
        hosts = set()
        for info in self.dispatchers.originalData.values():
            hosts.add(info.host.name)
        return list(hosts)

    async def dispose(self, quiet=False):
        if self._sendStatusTask:
            self._sendStatusTask.cancel()
        if self._checkStatusTask:
            self._checkStatusTask.cancel()
        if not quiet:
            await self._communicator.emit('aurevoir', DottedDict({'dispatcherId': self.id}))

    def __del__(self):
        EXECUTOR.callParallel(self.dispose, target=self)



from .dottedDict import DottedDict


def determineDifference(iter01, iter02):
    """ Determines the difference between the two sets.

    Args:
        iter01 (iterable): original set
        iter02 (iterable): adapted set

    Returns:
        DottedDict: containing `added` for the added elements and `removed` for the removed items.
    """
    added = list()
    removed = list()
    for item in iter01:
        if item not in iter02:
            removed.append(item)
    for item in iter02:
        if item not in iter01:
            added.append(item)
    return DottedDict({'added': added, 'removed': removed})


def union(set01: set, set02: set):
    return set01 | set02


def difference(set01: set, set02: set):
    return set01 - set02



from asyncio import sleep

import nope
from nope import getLayer, EXECUTOR, getDispatcher
from nope.plugins import install

"""
This is an Example how to extend the behavior of any class inside of Nope.

Therefore we want to load our sample plugin: "hello".
Therefore we know
"""

# install(nope, "nope.plugins.hello", plugin_dest="nope.dispatcher.rpcManager")
install(nope, "nope.plugins.hello")


async def main():
    dispatcher = getDispatcher({
        "communicator": await getLayer("event"),
        "logger": False,
    })

    manager = dispatcher.rpcManager

    async def hello_srv(greetings):
        print("Hello", greetings, "!")

    await dispatcher.ready.waitFor()
    await dispatcher.rpcManager.registerService(hello_srv, {"id": "hello_srv"})

    await sleep(0.1)

    manager.hello()

    await manager.performCall("hello_srv", ["reader"])

    await dispatcher.dispose()


EXECUTOR.loop.run_until_complete(main())



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from .baseModule import BaseModule
from ..helpers import ensureDottedAccess
from ..logger import getNopeLogger, DEBUG
from ..observable import NopeObservable
from ..eventEmitter import NopeEventEmitter


def _invertMode(options):
    _ret = options.copy()
    if isinstance(_ret.mode, list):
        if "subscribe" in _ret.mode:
            _ret.mode = ['publish', 'subscribe']
        else:
            _ret.mode = ['subscribe']
    elif _ret.mode == 'subscribe':
        _ret.mode = ['publish', 'subscribe']
    else:
        _ret.mode = ['subscribe']
    return _ret


class NopeGenericModule(BaseModule):
    def __init__(self, core):
        super().__init__(core)

        self._description = None
        self.dynamicInstanceMethods = dict()
        self.dynamicInstanceProperties = dict()
        self.dynamicInstanceEvents = dict()

    async def fromDescription(self, description, mode: str = 'overwrite'):
        """Initializes the Module, based on a Description.

        Args:
            description ([type]): [description]
            mode (str, optional): Defaults to 'overwrite'. Otherwise the data and Properties will be merged.

        Raises:
            Exception: [description]

        """

        # Adapt the Options:
        description = ensureDottedAccess(description)

        if mode == 'overwrite':
            await self.dispose()

            self.setAuthor(description.author.forename,
                           description.author.surename, description.author.mail)
            self.setVersion(description.version.version,
                            description.version.date)
            self.description = description.description
            self._type = description.type
            self.identifier = description.identifier

        if self.author is None:
            self._author = self._author = description.author

        if self.description is None:
            self.description = description.description

        if self.version is None:
            self._version = description.version

        if self.identifier is None:
            self.identifier = description.identifier
            self._logger = getNopeLogger(
                "generic-wrapper-" + self.identifier, DEBUG)

        for name, options in description.methods.items():

            self._logger.debug('Create function interface for "' + name + '"')

            async def _func(*args, _options=options):
                return await self._core.rpcManager.performCall(_options['id'], args, _options)

            if name in self.dynamicInstanceMethods:
                raise Exception(
                    'Name alread used. Not able to use the name twice')

            self.dynamicInstanceMethods[name] = _func

            # If the Function isnt dynamic, register it on the Object itself.
            if not options.get('isDynamic', False):
                if getattr(self, name, None) is not None:
                    raise Exception(
                        'Name alread used. Not able to use the name twice')

                setattr(self, name, _func)

            self._registeredMethods[name] = {
                'method': _func,
                'options': options
            }

        for name, options in description['properties'].items():

            self._logger.debug('Create property interface for "' + name + '"')

            # Add only elements, that are subscribed.
            # Properties, which are only publishing
            # should throw an error, if data is published
            # in a remote. self is done to maintain
            # consistency.
            if name in self.dynamicInstanceProperties:
                raise Exception(
                    'Name alread used. Not able to use the name twice')

            # Make shure it isnt published.
            # options.preventSendingToRegistery = true']

            # Register the Observable:
            self.dynamicInstanceProperties[name] = self._core.dataDistributor.register(
                # Assign a new Observable.
                NopeObservable(),
                # Use the provided Properties:
                _invertMode(options)
            )

            if not (options.get('isDynamic', False)):
                if getattr(self, name, None) is not None:
                    raise Exception(
                        'Name alread used. Not able to use the name twice')

                # Use the Same Element.
                setattr(self, name, self.dynamicInstanceProperties[name])

            self._registeredProperties[name] = {
                'observable': self.dynamicInstanceProperties[name],
                'options': options
            }

        for name, options in description['events'].items():

            self._logger.debug('Create event interface for "' + name + '"')

            # Add only elements, that are subscribed.
            # Properties, which are only publishing
            # should throw an error, if data is published
            # in a remote. self is done to maintain
            # consistency.
            if name in self.dynamicInstanceEvents:
                raise Exception(
                    'Name alread used. Not able to use the name twice')

            # Make shure it isnt published.
            # options.preventSendingToRegistery = true']

            # Register the Observable:
            self.dynamicInstanceEvents[name] = self._core.eventDistributor.register(
                # Assign a new Observable.
                NopeEventEmitter(),
                # Use the provided Properties:
                _invertMode(options)
            )

            if not (options.get('isDynamic', False)):
                if getattr(self, name, None) is not None:
                    raise Exception(
                        'Name alread used. Not able to use the name twice')

                # Use the Same Element.
                setattr(self, name, self.dynamicInstanceEvents[name])

            self._registeredEvents[name] = {
                'emitter': self.dynamicInstanceEvents[name],
                'options': options
            }

    async def registerProperty(self, *args):
        raise Exception('Function Should not be called on remote Instance!')

    async def registerMethod(self, *args):
        raise Exception('Function Should not be called on remote Instance!')

    async def unregisterMethod(self, *args):
        raise Exception('Function Should not be called on remote Instance!')

    async def unregisterProperty(self, *args):
        raise Exception('Function Should not be called on remote Instance!')

    async def init(self):
        try:
            await super().init()
        except BaseException:
            raise Exception('Call fromDescription before using')

    async def dispose(self):
        for name in self.dynamicInstanceProperties:
            self.dynamicInstanceProperties[name].dispose()
            # Remove Reference
            delattr(self, name)

        for name in self.dynamicInstanceEvents:
            self.dynamicInstanceEvents[name].dispose()
            # Remove Reference
            delattr(self, name)

        self.dynamicInstanceProperties = {}
        self.dynamicInstanceEvents = {}
        self.dynamicInstanceProperties = {}

        for name in self.dynamicInstanceMethods.keys():
            # Remove Reference
            delattr(self, name)

        self.dynamicInstanceMethods = {}
        self._registeredProperties = {}
        self._registeredEvents = {}
        self._registeredMethods = {}



from nope.helpers import ensureDottedAccess


def _class_decorator_factory(storing_type: str):
    def func(options):
        options = ensureDottedAccess(options)

        class class_decorator:
            """ A Simple Decorator for classes
                This will store the decorated element in an list
                named "decoratedItems"

                You can provide options etc.
            """

            def __init__(self, item):
                nonlocal storing_type
                self._type = storing_type

            def __set_name__(self, owner, name):
                # We want to store the item.
                if not hasattr(owner, "decoratedItems"):
                    setattr(owner, "decoratedItems", [])

                # Store item
                owner.decoratedItems.append(
                    ensureDottedAccess({
                        "accessor": name,
                        "options": options,
                        "type": self._type,
                        "owner": owner
                    })
                )

        return class_decorator

    return func


exportAsNopeService = _class_decorator_factory("method")
exportAsNopeProperty = _class_decorator_factory("property")
exportAsNopeEvent = _class_decorator_factory("event")



from collections.abc import Iterable

from .dottedDict import DottedDict, ensureDottedAccess
from .objectMethods import convertData, rgetattr, rqueryAttr, flattenObject, rsetattr
from .path import SPLITCHAR, getLeastCommonPathSegment, pathToCamelCase, pathToSnakeCase
from .prints import formatException
from .stringMethods import camelToSnake, snakeToCamel

__SENTINENTAL = object()


def keysToSnake(d: dict, adaptStrValues=False):
    """ helperfunction, which will change the keys to
        snake case.
    """
    ret = {}
    for key, value in d.items():
        if adaptStrValues and isinstance(value, str):
            ret[camelToSnake(key)] = camelToSnake(value)
        else:
            ret[camelToSnake(key)] = value

    return ret


def keysToSnakeNested(d: dict, adaptStrValues=False):
    """ helper function, which will change the keys to
        snake case. this will work with nested dicts as well
    """

    flatten = flattenObject(d, onlyPathToBaseValue=True)

    ret = ensureDottedAccess({})

    for k, v in flatten.items():
        if adaptStrValues and isinstance(v, str):
            rsetattr(ret, pathToSnakeCase(k), camelToSnake(v))
        else:
            rsetattr(ret, pathToSnakeCase(k), v)

    return ret


def keysToCamel(d: dict, adaptStrValues=False):
    """ helperfunction, which will change the options to
        camel case.
    """
    ret = {}
    for key, value in d.items():
        if adaptStrValues and isinstance(value, str):
            ret[snakeToCamel(key)] = snakeToCamel(value)
        else:
            ret[snakeToCamel(key)] = value

    return ret


def keysToCamelNested(d: dict, adaptStrValues=False):
    """ helperfunction, which will change the options to
        camel case. this will work with nested dicts as well
        (expecting every key contains a dict again.)
    """
    flatten = flattenObject(d, onlyPathToBaseValue=True)

    ret = ensureDottedAccess({})

    for k, v in flatten.items():
        if adaptStrValues and isinstance(v, str):
            rsetattr(ret, pathToCamelCase(k), snakeToCamel(v))
        else:
            rsetattr(ret, pathToCamelCase(k), v)

    return ret


def extractUniqueValues(d, path='', pathKey: str = None):
    """ Extracts the unique values of an map.

    Args:
        d (dict-like): The dict-like object that should be transformed.
        path (str, optional): The path of the attribute of that should be extracted of the value. Defaults to ''.
        pathKey (str, optional):   The Path of the unique key. If set to `None` -> The Item is selected directly.
                                    Defaults to None.

    Returns:
        set: A set containing the unique values.
    """
    if pathKey is None:
        pathKey = path
    if path != pathKey:
        _commonSegment = getLeastCommonPathSegment([path, pathKey], DottedDict(
            {'considerSingleLevel': False, 'considerMultiLevel': False}))

        # Assign the Value
        _commonSegment = _commonSegment if _commonSegment is not False else ""

        # Determine the Length of the common segment, to determine the relative
        # pathes
        _commonSegmentLength = len(_commonSegment.split(
            SPLITCHAR)) if _commonSegment != "" else 0

        _relPathContent = SPLITCHAR.join(
            path.split(SPLITCHAR)[_commonSegmentLength:])
        _relPathKey = SPLITCHAR.join(pathKey.split(SPLITCHAR)[
                                     _commonSegmentLength:])
        _items = extractValues(d, _commonSegment)
        _keysToUse = set()
        ret = []
        for item in _items:
            # Extract key and data:
            key = rgetattr(item, _relPathKey) if _relPathKey else item
            data = rgetattr(item, _relPathContent) if _relPathContent else item
            if key not in _keysToUse:
                # Only if the Key has not been used we are allowed to add
                # the data.
                _keysToUse.add(key)
                if data not in ret:
                    ret.append(data)
        return ret
    return extractValues(d, path, unique=True)


def extractValues(d, path='', unique=False):
    """ Helper to extract values of the map. Therefore the path must be provided.

    Example:
        >> d = {1:{"a": 0}, 2:{"a": 1}}
        >> r = extractValues(d,"a")

        r = [0,1]

    Args:
        d (dict-like): The dict-like object that should be transformed.
        path (str, optional): The path of the attribute of that should be extracted of the value. Defaults to ''.
        unique (bool, optional): Performs a check to extract only unique values.

    Returns:
        list: containing the element.
    """
    s = list()
    for v in d.values():
        if path:
            for item in rqueryAttr(v, path):
                if not unique or (unique and (item.data not in s)):
                    s.append(item.data)
        elif not unique or (unique and (v not in s)):
            s.append(v)
    return s


def transformDict(d, pathExtractedValue: str,
                  pathExtractedKey: str, logger=None):
    """_summary_

    Args:
        d (_type_): _description_
        pathExtractedValue (str): _description_
        pathExtractedKey (str): _description_

    Returns:
        _type_: _description_
    """

    # Initialize all values:
    keyMapping = dict()
    reverseKeyMapping = dict()
    conflicts = dict()
    extractedDict = dict()
    orgKeyToExtractedValue = dict()
    amountOf = dict()
    props = []
    onlyValidProps = True

    if isinstance(pathExtractedKey, str):
        props.append(DottedDict({'key': 'key', 'query': pathExtractedKey}))
        onlyValidProps = onlyValidProps and (len(pathExtractedKey) > 0)
    else:
        onlyValidProps = False

    if isinstance(pathExtractedValue, str):
        props.append(DottedDict(
            {'key': 'value', 'query': pathExtractedValue}))
        onlyValidProps = onlyValidProps and (len(pathExtractedValue) > 0)
    else:
        onlyValidProps = False

    keyIsHashable = True
    valueIsHashable = True

    __warned = False

    # Iterate over the items of the dict,
    # then we will extract the data stored in the Value.
    for k, v in d.items():
        extracted = []

        if onlyValidProps:
            extracted = convertData(v, props)

            for element in extracted:
                # Try to convert the data:
                try:
                    hash(element.key)
                    element.keyIsHashable = True
                except BaseException:
                    element.keyIsHashable = False
                    keyIsHashable = False

                # Try to convert the data:
                try:
                    hash(element.value)
                    element.valueIsHashable = True
                except BaseException:
                    element.valueIsHashable = False
                    valueIsHashable = False

        else:

            keys = []
            values = []

            # We migt adapt the key and the Value. Therefore we will use
            # the next if statements

            if isinstance(pathExtractedKey, str):
                if len(pathExtractedKey) > 0:
                    keys = rqueryAttr(v, pathExtractedKey)
                else:
                    keys = [v]
            else:
                keys = [k]

            if isinstance(pathExtractedValue, str):
                if len(pathExtractedValue) > 0:
                    values = rqueryAttr(v, pathExtractedValue)
                else:
                    values = [v]
            else:
                values = [v]

            for key in keys:
                for value in values:
                    data = DottedDict({
                        'key': key,
                        'value': value,
                        'keyIsHashable': True,
                        'valueIsHashable': True
                    })

                    # Try to convert the data:
                    try:
                        hash(data.key)
                    except BaseException:
                        data.keyIsHashable = False
                        keyIsHashable = False

                    # Try to convert the data:
                    try:
                        hash(data.value)
                    except BaseException:
                        data.valueIsHashable = False
                        valueIsHashable = False

                    extracted.append(data)

        # Create the entries for the following dicts.
        keyMapping[k] = set()
        orgKeyToExtractedValue[k] = set() if valueIsHashable else list()

        for item in extracted:

            if not item.keyIsHashable:
                error = Exception(
                    f"Can not hash the new key='{item.key}' (type={type(item.key)}) from path='{pathExtractedKey}'")

                if logger:
                    logger.error(error)
                else:
                    print(formatException(error))

                raise error

            if item.key in extractedDict:
                # If the extracted new key has already been defined,
                # we have to determine whether the stored item matches
                # the allready provided definition.
                if not (extractedDict.get(item.key) == item.value):
                    # Conflict detected -> Store it

                    if item.key not in conflicts:
                        conflicts[item.key] = set(
                        ) if valueIsHashable else list()

                    getattr(conflicts.get(item.key),
                            "add" if valueIsHashable else "append")(item.value)
                    getattr(conflicts.get(item.key), "add" if valueIsHashable else "append")(
                        extractedDict.get(item.key))
                else:
                    # No conflict -> just store the amount
                    amountOf[item.key] = amountOf.get(item.key, 0) + 1
            else:
                # Store the item and amount
                extractedDict[item.key] = item.value
                amountOf[item.key] = amountOf.get(item.key, 0) + 1

            #  If the reverse haven't been set ==> create it.
            if not (item.key in reverseKeyMapping):
                reverseKeyMapping[item.key] = set()

            # Store the mapping of new-key --> org-key.
            reverseKeyMapping[item.key].add(k)

            # Store the mapping of org-key --> new-key.
            keyMapping[k].add(item.key)

            # Now store the item.
            getattr(orgKeyToExtractedValue[k], "add" if valueIsHashable else "append")(
                item.value)

    return DottedDict({
        'extracted_map': extractedDict,
        'keyMapping': keyMapping,
        'conflicts': conflicts,
        'keyMappingreverse': reverseKeyMapping,
        'orgKeyToExtractedValue': orgKeyToExtractedValue,
        'amountOf': amountOf
    })


def reverse(d, path='', pathKey: str = None):
    """ Reverses the given map.

        If the path is provided, the Data is extracted based on the given path.
        If the `pathKey`, a different Key is used.

    Args:
        d (dict): The dict used as source.
        path (str, optional): The path for the attributes in the value. Defaults to ''.
        pathKey (str, optional): The path of the key, which will be provided in the `value`. Defaults to None.

    Returns:
        dict: The reversed Dict.
    """
    ret = dict()
    if pathKey is None:
        pathKey = path
    for k, v in d.items():
        keyToUse = k

        if pathKey:
            keyToUse = rgetattr(v, pathKey, __SENTINENTAL)
        valueToUse = v
        if path:
            valueToUse = rgetattr(v, path, __SENTINENTAL)
        if isinstance(valueToUse, Iterable):
            for _v in valueToUse:
                if _v not in ret:
                    ret[_v] = set()
                ret.get(_v).add(keyToUse)
        else:
            if valueToUse not in ret:
                ret[valueToUse] = set()
            ret.get(valueToUse).add(keyToUse)
    return ret



import asyncio
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial

from .prints import formatException


def isAsyncFunction(func) -> bool:
    """ Test whether the function is defined async or not.

    Args:
        func (_type_): _description_

    Returns:
        bool: _description_
    """
    return asyncio.iscoroutinefunction(func)


def getOrCreateEventloop():
    """ Creates or gets the default Eventloop.

    Returns:
        EventLoop: The determine Eventloop
    """
    try:
        return asyncio.get_event_loop()
    except RuntimeError as ex:
        if "There is no current event loop in thread" in str(ex):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return asyncio.get_event_loop()
        else:
            raise ex


class NopeExecutor:
    """ Helper for async execution. The executor helps you to run sync functions in extra threads
        in a compatible manner to NoPE.
    """

    def __init__(self, loop: asyncio.AbstractEventLoop = None,
                 executor: ThreadPoolExecutor | ProcessPoolExecutor = None):
        self._loop: asyncio.AbstractEventLoop = loop
        self._executor: ThreadPoolExecutor | ProcessPoolExecutor = executor
        if self._loop is None:
            self._loop = getOrCreateEventloop()

        self.logger = None
        self._todos = set()
        self._running = False

    def assignLoop(self, loop, forceDefaultToBeDefaultLoop=False):
        """ Helper to assign a event loop.

        Args:
            loop (_type_): _description_
            forceDefaultToBeDefaultLoop (bool, optional): _description_. Defaults to False.
        """
        self._loop = loop

        if self.logger:
            self.logger.warn("Use this function with care!")

        if forceDefaultToBeDefaultLoop:
            asyncio.set_event_loop(self._loop)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """ Returns the used loop

        Returns:
            asyncio.AbstractEventLoop: The internally used loop
        """
        return self._loop

    def useThreadPool(self, max_workers=None):
        self._executor = ThreadPoolExecutor(max_workers=max_workers)

    def useMultiProcessPool(self, max_workers=None):
        self._executor = ProcessPoolExecutor(max_workers=max_workers)

    def dispose(self, wait=True, cancel_futures=False):
        """ Disposes the Executor """
        if self._executor:
            self._executor.shutdown(wait=wait, cancel_futures=cancel_futures)

        self.loop.stop()
        self.loop.close()

    def run(self):
        """ Starts the eventloop and runs it forever """
        try:
            if self._running == False:
                self._running = True
                self.loop.run_forever()
        except KeyboardInterrupt:
            self.dispose()

    def generatePromise(self, **kwargs):
        """ Helper to create an Future.

            The keys and values of the future are df

        Returns:
            asyncio.Future: A Future
        """

        future = self.loop.create_future()

        def _cancel():
            pass

        setattr(future, 'cancelCallback', _cancel)

        for k, v in kwargs.items():
            if not hasattr(future, k):
                setattr(future, k, v)
            else:
                raise Exception(
                    f"You cant use the name {k}. It is predefined by the original Future.")

        return self.ensureExecution(future)

    def ensureExecution(self, todo):
        """ Make shure, the future / task is being executed. The user do not need to store the
            item to prevent the garbage collector to remove it.

        Args:
            todo (Future | Task): The item to take care of

        Returns:
            Future | Task: The item.
        """
        if isinstance(todo, (asyncio.Task, asyncio.Future)):
            self._todos.add(todo)

            def remove(*args, **kwargs):
                self._todos.remove(todo)

            todo.add_done_callback(remove)

        return todo

    def setTimeout(self, func, timeout_ms: int, *args, **kwargs):
        """ Function to call the function after the delay.

        Args:
            func (function): The function to call.
            timeout_ms (int): Delay in ms.

        Returns:
            asyncio.Task: The task, executing the timeout.
        """
        function_to_use = self.wrapFuncIfRequired(func)

        async def timeout():
            promise = None
            try:
                await asyncio.sleep(timeout_ms / 1000.0)
                promise = function_to_use(*args, **kwargs)
                await promise
            except Exception as error:
                if promise:
                    if self.logger:
                        self.logger.error(
                            "Exception raised during executing 'setTimeout'")
                        self.logger.error(formatException(error))
                    else:
                        print(formatException(error))

        task = self.loop.create_task(timeout())

        return self.ensureExecution(task)

    def setInterval(self, func, interval_ms: int, *args,
                    **kwargs) -> asyncio.Task:
        """ Creates an interval which will be called

        Args:
            func (function): The function to call.
            interval_ms (int): Interval in ms.

        Returns:
            asyncio.Task: The task, executing the interval.
        """
        function_to_use = self.wrapFuncIfRequired(func)

        async def interval():
            try:
                promise = None
                while True:
                    await asyncio.sleep(interval_ms / 1000.0)
                    promise = function_to_use(*args, **kwargs)
                    await promise
            except Exception as error:
                if self.logger:
                    self.logger.error(
                        "Exception raised during executing 'interval'")
                    self.logger.error(formatException(error))
                else:
                    print(formatException(error))

        task = self.loop.create_task(interval())

        return self.ensureExecution(task)

    def callParallel(self, func, *args, done_callback=None, **
                     kwargs) -> asyncio.Task | asyncio.Future:
        """ Helper, which will call the function in the Background.


        Args:
            func (function): The function, which should be called in parallel

        Returns:
            asyncio.Task | asyncio.Future: The Task for the Function.
        """
        function_to_use = self.wrapFuncIfRequired(func)
        task = self.loop.create_task(function_to_use(*args, **kwargs))

        def on_done(future: asyncio.Task):
            err = future.exception()
            if err:
                if self.logger:
                    self.logger.error(err)
                    self.logger.error(formatException(err))
                else:
                    print(err)
                    print(formatException(err))

            elif done_callback:
                done_callback(future)

        # if done_callback:
        #     task.add_done_callback(done_callback)
        # else:
        task.add_done_callback(on_done)

        return self.ensureExecution(task)

    def wrapFuncIfRequired(self, func):
        """ Helper to asyncify a the provided function if requried. If it is already a coroutine,
            we will return this, otherwise the function will be wrapped into an async function.

        Args:
            func (function): The function, which must be will be wrapped

        Returns:
            corutine: The function as async function
        """
        if not callable(func):
            raise TypeError("The parameter 'func' is not callable")

        if not asyncio.iscoroutinefunction(func):
            async def run(*args, **kwargs):
                pfunc = partial(func, *args, **kwargs)
                try:
                    return await self.loop.run_in_executor(self._executor, pfunc)
                except Exception as error:
                    if self.logger:
                        self.logger.error(
                            "Exception raised during executing a wrapped sync method '_wrapFuncIfRequired'")
                        self.logger.error(formatException(error))
                    else:
                        print(formatException(error))

            return run
        else:
            return func


EXECUTOR = NopeExecutor()
EXECUTOR.useThreadPool()


def Promise(callback):
    """ Creates a NodeJS like Promise

    Args:
        callback (function or coroutine): The function receiving resolve and reject. Best to be sync.

    Returns:
        Future: The awaitable Future.
    """
    future = EXECUTOR.generatePromise()

    def reject(error):
        if future.done():
            print(formatException(Exception("Called 'reject' multiple times")))
            return
        future.set_exception(error)

    def resolve(value):
        if future.done():
            print(formatException(Exception("Called 'resolve' multiple times")))
            return
        future.set_result(value)

    # Now we want call the callback in an extra thread.
    callback(resolve, reject)

    return future


async def waitFor(callback, initalWaitInMs: int = None, testFirst: bool = True, maxRetries: int = float("inf"),
                  delay: int = 50, maxTimeout: int = None, additionalDelay: int = None):
    """ Function which will wait for the callback to return true

    Args:
        callback (function): The Callback which will be checkt with the given delay. The Function must return a boolean value. It can be a coroutine.
        initalWaitInMs (int, optional): First amount of time to wait. Defaults to None.
        testFirst (bool, optional): Flag to enable Testing directly (after the intial wait time). Defaults to True.
        maxRetries (int, optional): Number of allowed retries. Defaults to float("inf").
        delay (int, optional): The delay in *ms* after which the elements are tested again. Defaults to 50 [ms].
        maxTimeout (int, optional): The max Time to wait. Throws an error if reached. Defaults to None.
        additionalDelay (int, optional): _description_. Defaults to None.

    Returns:
        Promise: A Promise which can be awaited.
    """

    # Ensure our callback is async.
    asyncCallback = EXECUTOR.wrapFuncIfRequired(callback)

    # Define our callback, that we will hand over to our Promise.

    async def cb(resolve, reject):

        if initalWaitInMs:
            # We will wait
            await asyncio.sleep(initalWaitInMs / 1000.0)

        try:
            if testFirst and await asyncCallback():
                if additionalDelay:
                    await asyncio.sleep(additionalDelay / 1000.0)

                resolve(True)
                return
            else:
                retryCounter = 0

                timeout: asyncio.Task = None
                interval: asyncio.Task = None

                # If there is a Timeout, define a Timeout Function, which will
                # Throw an Error on Timeout.

                if maxTimeout:
                    def onTimeout():
                        if interval:
                            interval.cancel()

                        reject(TimeoutError("Timed out!"))

                    timeout = EXECUTOR.setTimeout(onTimeout, maxTimeout)

                # Define a Testfunction, which will periodically test whether the condition is
                # fullfield or not. Internally it counts the number of retries, if the max allowed
                # number of retries has been reached => Throw an Error

                async def inInterval():
                    try:
                        nonlocal retryCounter
                        if maxRetries and maxRetries > retryCounter:
                            # Stop the interval
                            interval.cancel()

                            if timeout:
                                timeout.cancel()

                            reject(Exception("Max Retries has been reached!"))
                        elif await asyncCallback():
                            # Stop the interval
                            interval.cancel()

                            if timeout:
                                timeout.cancel()

                            if additionalDelay:
                                await asyncio.sleep(additionalDelay / 1000.0)

                            resolve(True)

                        retryCounter += 1

                    except Exception as e:
                        # Stop the interval
                        interval.cancel()

                        if timeout:
                            timeout.cancel()

                        # Forward the Error.
                        reject(e)

                interval = EXECUTOR.setInterval(inInterval, delay)

        except Exception as e:
            reject(e)

    return Promise(cb)



from .path import SPLITCHAR


def getPropertyPath(identifier, name):
    return f'{identifier}{SPLITCHAR}properties{SPLITCHAR}{name}'


def isPropertyPathCorrect(identifier, path):
    return path.startswith(f'{identifier}{SPLITCHAR}properties{SPLITCHAR}')


def getMethodPath(identifier, name):
    return f'{identifier}{SPLITCHAR}methods{SPLITCHAR}{name}'


def isMethodPathCorrect(identifier, path):
    return path.startswith(f'{identifier}{SPLITCHAR}methods{SPLITCHAR}')


def getEmitterPath(identifier, name):
    return f'{identifier}{SPLITCHAR}events{SPLITCHAR}{name}'


def isEmitterPathCorrect(identifier, path):
    return path.startswith(f'{identifier}{SPLITCHAR}events{SPLITCHAR}')



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import importlib
import sys


def dynamicImport(name: str, pathToFile: str):
    """ Helper to dynamically import content of a file.

    Args:
        name (str): name of the object to load
        pathToFile (str): the path to the file.

    Returns:
        any: the loaded module
    """
    spec = importlib.util.spec_from_file_location(
        name, pathToFile)

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)

    return module



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from .baseModule import BaseModule
from .genericModule import NopeGenericModule

__all__ = ["BaseModule", "NopeGenericModule"]



#!/usr/bin/env python

# @author Martin Karkowski

# @email m.karkowski@zema.de


from . import (asyncHelpers, dictMethods, dispatcherPathes, dottedDict,

               emitter, idMethods, importing, objectMethods, path,

               pathMatchingMethods, prints, runtime, stringMethods, timers,

               timestamp, hashable, listMethods, jsonMethods, files)

from .asyncHelpers import (

    Promise,

    getOrCreateEventloop,

    isAsyncFunction,

    waitFor,

    EXECUTOR)

from .dictMethods import (extractUniqueValues, extractValues,

                          keysToCamel, keysToCamelNested, keysToSnake,

                          keysToSnakeNested, transformDict)
from .dispatcherPathes import (getEmitterPath, getMethodPath,

                               getPropertyPath, isEmitterPathCorrect,

                               isMethodPathCorrect,

                               isPropertyPathCorrect)
from .dottedDict import DottedDict, convertToDottedDict, ensureDottedAccess
from .emitter import Emitter
from .files import createFile
from .hashable import hlist, hset, hdict
from .idMethods import generateId
from .importing import dynamicImport
from .jsonMethods import dumps, loads
from .listMethods import (avgOfArray, extractListElements, isIterable, isList, flattenDeep,
                          maxOfArray, minOfArray)
from .objectMethods import (convertData, copy, deepAssign, deflattenObject,

                            flattenObject, getKeys, isFloat, isInt,

                            isNumber, isDictLike,

                            keepProperties, objectToDict,

                            recursiveForEach, rgetattr, rqueryAttr, rsetattr)
from .path import (
    MULTI_LEVEL_WILDCARD,
    SINGLE_LEVEL_WILDCARD,
    SPLITCHAR,
    containsWildcards,
    convertPath,
    pathToCamelCase,
    pathToSnakeCase,
    getLeastCommonPathSegment,
    patternIsValid,
    toPythonPath,
    varifyPath)
from .pathMatchingMethods import comparePatternAndPath
from .prints import formatException
from .runtime import offload_function_to_thread
from .setMethods import determineDifference, difference, union
from .stringMethods import camelToSnake, insertNewLines, insert, limitString, padString, replaceAll, snakeToCamel, toCamelCase, toSnakeCase, toVariableName
from .timers import setInterval, setTimeout
from .timestamp import getTimestamp



from nope.communication.bridge import Bridge
from nope.communication.layers import MQTTLayer, IoSocketClientLayer
from nope.helpers import generateId, DottedDict

VALID_LAYERS = DottedDict({
    'event': Bridge,
    'io-client': IoSocketClientLayer,
    'mqtt': MQTTLayer
})

LAYER_DEFAULT_PARAMETERS = DottedDict({
    'io-client': 'http://127.0.0.1:7000',
    'mqtt': 'mqtt://localhost:1883'
})


async def getLayer(layer: str, parameter=None, logger=False):
    bridge = Bridge(generateId(), logger)
    params = parameter if parameter is not None else LAYER_DEFAULT_PARAMETERS.get(
        layer, False)
    if layer == 'event':
        pass
    elif layer == 'io-client':
        await bridge.addCommunicationLayer(IoSocketClientLayer(params, logger), True)
    elif layer == 'mqtt':
        await bridge.addCommunicationLayer(MQTTLayer(params, logger))

    bridge.connected.forcePublish()
    return bridge
